"""

compare_batch_results.py

Compare sets of batch results; typically used to compare:

* Results from different MegaDetector versions
* Results before/after RDE
* Results with/without augmentation

Makes pairwise comparisons, but can take lists of results files (will perform 
all pairwise comparisons).  Results are written to an HTML page that shows the number
and nature of disagreements (in the sense of each image being a detection or non-detection), 
with sample images for each category.

"""

#%% Imports

import json
import os
import random
import copy
import urllib
import itertools

from tqdm import tqdm
from functools import partial

from multiprocessing.pool import ThreadPool
from multiprocessing.pool import Pool

from megadetector.visualization import visualization_utils
from megadetector.utils.write_html_image_list import write_html_image_list
from megadetector.utils import path_utils


#%% Constants and support classes
    
class PairwiseBatchComparisonOptions:
    """
    Defines the options used for a single pairwise comparison; a list of these
    pairwise options sets is stored in the BatchComparisonsOptions class.
    """
    
    def __init__(self):
        
        #: First filename to compare
        self.results_filename_a = None
        
        #: Second filename to compare
        self.results_filename_b = None
        
        #: Description to use in the output HTML for filename A
        self.results_description_a = None
        
        #: Description to use in the output HTML for filename B
        self.results_description_b = None
        
        #: Per-class detection thresholds to use for filename A (including a 'default' threshold)
        self.detection_thresholds_a = {'animal':0.15,'person':0.15,'vehicle':0.15,'default':0.15}
        
        #: Per-class detection thresholds to use for filename B (including a 'default' threshold)
        self.detection_thresholds_b = {'animal':0.15,'person':0.15,'vehicle':0.15,'default':0.15}
    
        #: Rendering threshold to use for all categories for filename A
        self.rendering_confidence_threshold_a = 0.1
        
        #: Rendering threshold to use for all categories for filename B
        self.rendering_confidence_threshold_b = 0.1

# ...class PairwiseBatchComparisonOptions


class BatchComparisonOptions:
    """
    Defines the options for a set of (possibly many) pairwise comparisons.
    """
    
    def __init__(self):
        
        #: Folder to which we should write HTML output
        self.output_folder = None
        
        #: Base folder for images (which are specified as relative files)
        self.image_folder = None
        
        #: Job name to use in the HTML output file
        self.job_name = ''
        
        #: Maximum number of images to render for each category, where a "category" here is
        #: "detections_a_only", "detections_b_only", etc., or None to render all images.
        self.max_images_per_category = 1000
        
        #: Maximum number of images per HTML page (paginates if a category page goes beyond this),
        #: or None to disable pagination.
        self.max_images_per_page = None
        
        #: Colormap to use for detections in file A (maps detection categories to colors)
        self.colormap_a = ['Red']
        
        #: Colormap to use for detections in file B (maps detection categories to colors)
        self.colormap_b = ['RoyalBlue']
    
        #: Process-based parallelization isn't supported yet; this must be "True"
        self.parallelize_rendering_with_threads = True
        
        #: List of filenames to include in the comparison, or None to use all files
        self.filenames_to_include = None
        
        #: Compare only detections/non-detections, ignore categories (still renders categories)
        self.class_agnostic_comparison = False
        
        #: Width of images to render in the output HTML
        self.target_width = 800
        
        #: Number of workers to use for rendering, or <=1 to disable parallelization
        self.n_rendering_workers = 20
        
        #: Random seed for image sampling (not used if max_images_per_category is None)
        self.random_seed = 0
        
        #: Whether to sort results by confidence; if this is False, sorts by filename
        self.sort_by_confidence = False
        
        #: The expectation is that all results sets being compared will refer to the same images; if this
        #: is True (default), we'll error if that's not the case, otherwise non-matching lists will just be
        #: a warning.
        self.error_on_non_matching_lists = True
        
        #: List of PairwiseBatchComparisonOptions that defines the comparisons we'll render.
        self.pairwise_options = []
        
# ...class BatchComparisonOptions
    

class PairwiseBatchComparisonResults:
    """
    The results from a single pairwise comparison.
    """
    
    def __init__(self):
    
        #: String of HTML content suitable for rendering to an HTML file
        self.html_content = None
        
        #: Possibly-modified version of the PairwiseBatchComparisonOptions supplied as input.
        self.pairwise_options = None
        
        #: A dictionary with keys including:
        #:
        #: common_detections
        #: common_non_detections
        #: detections_a_only
        #: detections_b_only
        #: class_transitions
        #
        #: Each of these maps a filename to a two-element list (the image in set A, the image in set B).
        self.categories_to_image_pairs = None

# ...class PairwiseBatchComparisonResults
    
    
class BatchComparisonResults:
    """
    The results from a set of pairwise comparisons
    """
    
    def __init__(self):
        
        #: Filename containing HTML output
        self.html_output_file = None
        
        #: A list of PairwiseBatchComparisonResults
        self.pairwise_results = None
    
# ...class BatchComparisonResults    


main_page_style_header = """<head>
    <style type="text/css">
    a { text-decoration: none; }
    body { font-family: segoe ui, calibri, "trebuchet ms", verdana, arial, sans-serif; }
    div.contentdiv { margin-left: 20px; }
    </style>
    </head>"""

main_page_header = '<html>\n{}\n<body>\n'.format(main_page_style_header)
main_page_footer = '<br/><br/><br/></body></html>\n'


#%% Comparison functions

def _render_image_pair(fn,image_pairs,category_folder,options,pairwise_options):
    """
    Render two sets of results (i.e., a comparison) for a single image.
    
    Args:
        fn (str): image filename
        image_pairs (dict): dict mapping filenames to pairs of image dicts
        category_folder (str): folder to which to render this image, typically 
            "detections_a_only", "detections_b_only", etc.
        options (BatchComparisonOptions): job options
        pairwise_options (PairwiseBatchComparisonOptions): pairwise comparison options
            
    Returns:
        str: rendered image filename            
    """
    
    input_image_path = os.path.join(options.image_folder,fn)
    assert os.path.isfile(input_image_path), 'Image {} does not exist'.format(input_image_path)
    
    im = visualization_utils.open_image(input_image_path)
    image_pair = image_pairs[fn]
    detections_a = image_pair[0]['detections']
    detections_b = image_pair[1]['detections']
    
    custom_strings_a = [''] * len(detections_a)
    custom_strings_b = [''] * len(detections_b)    
    
    # This function is often used to compare results before/after various merging
    # steps, so we have some special-case formatting based on the "transferred_from"
    # field generated in merge_detections.py.
    for i_det,det in enumerate(detections_a):
        if 'transferred_from' in det:
            custom_strings_a[i_det] = '({})'.format(
                det['transferred_from'].split('.')[0])
    
    for i_det,det in enumerate(detections_b):
        if 'transferred_from' in det:
            custom_strings_b[i_det] = '({})'.format(
                det['transferred_from'].split('.')[0])
    
    if options.target_width is not None:
        im = visualization_utils.resize_image(im, options.target_width)
        
    visualization_utils.render_detection_bounding_boxes(detections_a,im,
        confidence_threshold=pairwise_options.rendering_confidence_threshold_a,
        thickness=4,expansion=0,
        colormap=options.colormap_a,
        textalign=visualization_utils.TEXTALIGN_LEFT,
        custom_strings=custom_strings_a)
    visualization_utils.render_detection_bounding_boxes(detections_b,im,
        confidence_threshold=pairwise_options.rendering_confidence_threshold_b,
        thickness=2,expansion=0,
        colormap=options.colormap_b,
        textalign=visualization_utils.TEXTALIGN_RIGHT,
        custom_strings=custom_strings_b)

    output_image_fn = path_utils.flatten_path(fn)
    output_image_path = os.path.join(category_folder,output_image_fn)
    im.save(output_image_path)
    return output_image_path

# ...def _render_image_pair()


def _pairwise_compare_batch_results(options,output_index,pairwise_options):
    """
    The main entry point for this module is compare_batch_results(), which calls 
    this function for each pair of comparisons the caller has requested.  Generates an
    HTML page for this comparison.  Returns a BatchComparisonResults object.
    
    Args:
        options (BatchComparisonOptions): overall job options for this comparison group
        output_index (int): a numeric index used for generating HTML titles    
        pairwise_options (PairwiseBatchComparisonOptions): job options for this comparison
        
    Returns:
        PairwiseBatchComparisonResults: the results of this pairwise comparison
    """
    
    # pairwise_options is passed as a parameter here, and should not be specified
    # in the options object.
    assert options.pairwise_options is None
    
    if options.random_seed is not None:
        random.seed(options.random_seed)

    # Warn the user if some "detections" might not get rendered
    max_classification_threshold_a = max(list(pairwise_options.detection_thresholds_a.values()))
    max_classification_threshold_b = max(list(pairwise_options.detection_thresholds_b.values()))
    
    if pairwise_options.rendering_confidence_threshold_a > max_classification_threshold_a:
        print('*** Warning: rendering threshold A ({}) is higher than max confidence threshold A ({}) ***'.format(
            pairwise_options.rendering_confidence_threshold_a,max_classification_threshold_a))
    
    if pairwise_options.rendering_confidence_threshold_b > max_classification_threshold_b:
        print('*** Warning: rendering threshold B ({}) is higher than max confidence threshold B ({}) ***'.format(
            pairwise_options.rendering_confidence_threshold_b,max_classification_threshold_b))
    

    ##%% Validate inputs
    
    assert os.path.isfile(pairwise_options.results_filename_a), \
        "Can't find results file {}".format(pairwise_options.results_filename_a)
    assert os.path.isfile(pairwise_options.results_filename_b), \
        "Can't find results file {}".format(pairwise_options.results_filename_b)
    assert os.path.isdir(options.image_folder), \
        "Can't find image folder {}".format(pairwise_options.image_folder)
    os.makedirs(options.output_folder,exist_ok=True)
    
    
    ##%% Load both result sets
    
    with open(pairwise_options.results_filename_a,'r') as f:
        results_a = json.load(f)
    
    with open(pairwise_options.results_filename_b,'r') as f:
        results_b = json.load(f)
        
    # Don't let path separators confuse things
    for im in results_a['images']:
        if 'file' in im:
            im['file'] = im['file'].replace('\\','/')
    for im in results_b['images']:
        if 'file' in im:
            im['file'] = im['file'].replace('\\','/')
    
    if not options.class_agnostic_comparison:
        assert results_a['detection_categories'] == results_b['detection_categories'], \
            "Cannot perform a class-sensitive comparison across results with different categories"
            
    detection_categories_a = results_a['detection_categories']
    detection_categories_b = results_b['detection_categories']
    
    if pairwise_options.results_description_a is None:
        if 'detector' not in results_a['info']:
            print('No model metadata supplied for results-A, assuming MDv4')
            pairwise_options.results_description_a = 'MDv4 (assumed)'
        else:            
            pairwise_options.results_description_a = results_a['info']['detector']
    
    if pairwise_options.results_description_b is None:
        if 'detector' not in results_b['info']:
            print('No model metadata supplied for results-B, assuming MDv4')
            pairwise_options.results_description_b = 'MDv4 (assumed)'
        else:            
            pairwise_options.results_description_b = results_b['info']['detector']
    
    images_a = results_a['images']
    images_b = results_b['images']
    
    filename_to_image_a = {im['file']:im for im in images_a}
    filename_to_image_b = {im['file']:im for im in images_b}
    
    
    ##%% Make sure they represent the same set of images
    
    filenames_a = [im['file'] for im in images_a]
    filenames_b_set = set([im['file'] for im in images_b])
    
    if len(images_a) != len(images_b):
        s = 'set A has {} images, set B has {}'.format(len(images_a),len(images_b))
        if options.error_on_non_matching_lists:
            raise ValueError(s)
        else:
            print('Warning: ' + s)
    else:
        if options.error_on_non_matching_lists:
            for fn in filenames_a:
                assert fn in filenames_b_set

    assert len(filenames_a) == len(images_a)
    assert len(filenames_b_set) == len(images_b)    
    
    if options.filenames_to_include is None:
        filenames_to_compare = filenames_a
    else:
        filenames_to_compare = options.filenames_to_include
    
    ##%% Find differences
    
    # Each of these maps a filename to a two-element list (the image in set A, the image in set B)
    #
    # Right now, we only handle a very simple notion of class transition, where the detection
    # of maximum confidence changes class *and* both images have an above-threshold detection.
    common_detections = {}
    common_non_detections = {}
    detections_a_only = {}
    detections_b_only = {}
    class_transitions = {}
        
    # fn = filenames_to_compare[0]
    for fn in tqdm(filenames_to_compare):
    
        if fn not in filename_to_image_b:
            
            # We shouldn't have gotten this far if error_on_non_matching_lists is set
            assert not options.error_on_non_matching_lists
            
            print('Skipping filename {}, not in image set B'.format(fn))
            continue
        
        im_a = filename_to_image_a[fn]
        im_b = filename_to_image_b[fn]
        
        categories_above_threshold_a = set()

        if not 'detections' in im_a or im_a['detections'] is None:
            assert 'failure' in im_a and im_a['failure'] is not None
            continue
        
        if not 'detections' in im_b or im_b['detections'] is None:
            assert 'failure' in im_b and im_b['failure'] is not None
            continue
                
        invalid_category_error = False

        # det = im_a['detections'][0]
        for det in im_a['detections']:
            
            category_id = det['category']
            
            if category_id not in detection_categories_a:
                print('Warning: unexpected category {} for model A on file {}'.format(category_id,fn))
                invalid_category_error = True
                break
                
            conf = det['conf']
            
            if detection_categories_a[category_id] in pairwise_options.detection_thresholds_a:
                conf_thresh = pairwise_options.detection_thresholds_a[detection_categories_a[category_id]]
            else:
                conf_thresh = pairwise_options.detection_thresholds_a['default']
                
            if conf >= conf_thresh:
                categories_above_threshold_a.add(category_id)
                            
        if invalid_category_error:
            continue
        
        categories_above_threshold_b = set()
        
        for det in im_b['detections']:
            
            category_id = det['category']
            
            if category_id not in detection_categories_b:
                print('Warning: unexpected category {} for model B on file {}'.format(category_id,fn))
                invalid_category_error = True
                break
            
            conf = det['conf']
            
            if detection_categories_b[category_id] in pairwise_options.detection_thresholds_b:
                conf_thresh = pairwise_options.detection_thresholds_b[detection_categories_b[category_id]]
            else:
                conf_thresh = pairwise_options.detection_thresholds_a['default']
                
            if conf >= conf_thresh:
                categories_above_threshold_b.add(category_id)
                            
        if invalid_category_error:
            continue
        
        im_pair = (im_a,im_b)
        
        detection_a = (len(categories_above_threshold_a) > 0)
        detection_b = (len(categories_above_threshold_b) > 0)
                
        if detection_a and detection_b:            
            if (categories_above_threshold_a == categories_above_threshold_b) or \
                options.class_agnostic_comparison:
                common_detections[fn] = im_pair
            else:
                class_transitions[fn] = im_pair
        elif (not detection_a) and (not detection_b):
            common_non_detections[fn] = im_pair
        elif detection_a and (not detection_b):
            detections_a_only[fn] = im_pair
        else:
            assert detection_b and (not detection_a)
            detections_b_only[fn] = im_pair
            
    # ...for each filename
    
    print('Of {} files:\n{} common detections\n{} common non-detections\n{} A only\n{} B only\n{} class transitions'.format(
        len(filenames_to_compare),len(common_detections),
        len(common_non_detections),len(detections_a_only),
        len(detections_b_only),len(class_transitions)))
        
    
    ##%% Sample and plot differences
    
    if options.n_rendering_workers > 1:
       worker_type = 'processes'
       if options.parallelize_rendering_with_threads:
           worker_type = 'threads'
       print('Rendering images with {} {}'.format(options.n_rendering_workers,worker_type))
       if options.parallelize_rendering_with_threads:
           pool = ThreadPool(options.n_rendering_workers)    
       else:
           pool = Pool(options.n_rendering_workers)    
        
    categories_to_image_pairs = {
        'common_detections':common_detections,
        'common_non_detections':common_non_detections,
        'detections_a_only':detections_a_only,
        'detections_b_only':detections_b_only,
        'class_transitions':class_transitions
    }
    
    categories_to_page_titles = {
        'common_detections':'Detections common to both models',
        'common_non_detections':'Non-detections common to both models',
        'detections_a_only':'Detections reported by model A only',
        'detections_b_only':'Detections reported by model B only',
        'class_transitions':'Detections reported as different classes by models A and B'
    }

    local_output_folder = os.path.join(options.output_folder,'cmp_' + \
                                       str(output_index).zfill(3))
    
    def render_detection_comparisons(category,image_pairs,image_filenames):
        
        print('Rendering detections for category {}'.format(category))
                
        category_folder = os.path.join(local_output_folder,category)
        os.makedirs(category_folder,exist_ok=True)
        
        # fn = image_filenames[0]
        if options.n_rendering_workers <= 1:
            output_image_paths = []
            for fn in tqdm(image_filenames):        
                output_image_paths.append(_render_image_pair(fn,image_pairs,category_folder,
                                                            options,pairwise_options))
        else:            
            output_image_paths = list(tqdm(pool.imap(
                partial(_render_image_pair, image_pairs=image_pairs, 
                        category_folder=category_folder,options=options,
                        pairwise_options=pairwise_options),
                image_filenames), 
                total=len(image_filenames)))
        
        return output_image_paths
    
    # ...def render_detection_comparisons()
    
    # For each category, generate comparison images and the 
    # comparison HTML page.
    #
    # category = 'common_detections'
    for category in categories_to_image_pairs.keys():
        
        # Choose detection pairs we're going to render for this category
        image_pairs = categories_to_image_pairs[category]
        image_filenames = list(image_pairs.keys())
        
        if options.max_images_per_category is not None and options.max_images_per_category > 0:
            if len(image_filenames) > options.max_images_per_category:
                print('Sampling {} of {} image pairs for category {}'.format(
                    options.max_images_per_category,
                    len(image_filenames),
                    category))
                image_filenames = random.sample(image_filenames,
                                                options.max_images_per_category)
            assert len(image_filenames) <= options.max_images_per_category

        input_image_absolute_paths = [os.path.join(options.image_folder,fn) for fn in image_filenames]
        
        category_image_output_paths = render_detection_comparisons(category,
                                                            image_pairs,image_filenames)
        
        category_html_filename = os.path.join(local_output_folder,
                                              category + '.html')
        category_image_output_paths_relative = [os.path.relpath(s,local_output_folder) \
                                         for s in category_image_output_paths]
        
        image_info = []
        
        assert len(category_image_output_paths_relative) == len(input_image_absolute_paths)
        
        for i_fn,fn in enumerate(category_image_output_paths_relative): 
            
            input_path_relative = image_filenames[i_fn]
            image_pair = image_pairs[input_path_relative]
            assert len(image_pair) == 2; image_a = image_pair[0]; image_b = image_pair[1]
            
            def maxempty(L):
                if len(L) == 0:
                    return 0
                else:
                    return max(L)
                
            max_conf_a = maxempty([det['conf'] for det in image_a['detections']])
            max_conf_b = maxempty([det['conf'] for det in image_b['detections']])
            
            title = input_path_relative + ' (max conf {:.2f},{:.2f})'.format(max_conf_a,max_conf_b)
            
            # Only used if sort_by_confidence is True
            if category == 'common_detections':
                sort_conf = max(max_conf_a,max_conf_b)
            elif category == 'common_non_detections':
                sort_conf = max(max_conf_a,max_conf_b)
            elif category == 'detections_a_only':
                sort_conf = max_conf_a
            elif category == 'detections_b_only':
                sort_conf = max_conf_b
            elif category == 'class_transitions':
                sort_conf = max(max_conf_a,max_conf_b)
            else:
                print('Warning: unknown sort category {}'.format(category))
                sort_conf = max(max_conf_a,max_conf_b)
                
            info = {
                'filename': fn,
                'title': title,
                'textStyle': 'font-family:verdana,arial,calibri;font-size:' + \
                    '80%;text-align:left;margin-top:20;margin-bottom:5',
                'linkTarget': urllib.parse.quote(input_image_absolute_paths[i_fn]),
                'sort_conf':sort_conf
            }
            image_info.append(info)
    
        # ...for each image
        
        category_page_header_string = '<h1>{}</h1>'.format(categories_to_page_titles[category])
        category_page_header_string += '<p style="font-weight:bold;">\n'
        category_page_header_string += 'Model A: {}<br/>\n'.format(
            pairwise_options.results_description_a)
        category_page_header_string += 'Model B: {}'.format(pairwise_options.results_description_b)
        category_page_header_string += '</p>\n'
        
        category_page_header_string += '<p>\n'
        category_page_header_string += 'Detection thresholds for A ({}):\n{}<br/>'.format(
            pairwise_options.results_description_a,str(pairwise_options.detection_thresholds_a))
        category_page_header_string += 'Detection thresholds for B ({}):\n{}<br/>'.format(
            pairwise_options.results_description_b,str(pairwise_options.detection_thresholds_b))
        category_page_header_string += 'Rendering threshold for A ({}):\n{}<br/>'.format(
            pairwise_options.results_description_a,
            str(pairwise_options.rendering_confidence_threshold_a))
        category_page_header_string += 'Rendering threshold for B ({}):\n{}<br/>'.format(
            pairwise_options.results_description_b,
            str(pairwise_options.rendering_confidence_threshold_b))
        category_page_header_string += '</p>\n'        
        
        # Default to sorting by filename
        if options.sort_by_confidence:
            image_info = sorted(image_info, key=lambda d: d['sort_conf'], reverse=True)
        else:
            image_info = sorted(image_info, key=lambda d: d['filename'])
            
        write_html_image_list(
            category_html_filename,
            images=image_info,
            options={
                'headerHtml': category_page_header_string,
                'maxFiguresPerHtmlFile': options.max_images_per_page
            })
        
    # ...for each category
    
    
    ##%% Write the top-level HTML file content

    html_output_string  = ''
    
    html_output_string += '<p>Comparing <b>{}</b> (A, red) to <b>{}</b> (B, blue)</p>'.format(
        pairwise_options.results_description_a,pairwise_options.results_description_b)
    html_output_string += '<div class="contentdiv">\n'
    html_output_string += 'Detection thresholds for {}:\n{}<br/>'.format(
        pairwise_options.results_description_a,
        str(pairwise_options.detection_thresholds_a))
    html_output_string += 'Detection thresholds for {}:\n{}<br/>'.format(
        pairwise_options.results_description_b,
        str(pairwise_options.detection_thresholds_b))
    html_output_string += 'Rendering threshold for {}:\n{}<br/>'.format(
        pairwise_options.results_description_a,
        str(pairwise_options.rendering_confidence_threshold_a))
    html_output_string += 'Rendering threshold for {}:\n{}<br/>'.format(
        pairwise_options.results_description_b,
        str(pairwise_options.rendering_confidence_threshold_b))
    
    html_output_string += '<br/>'
    
    html_output_string += 'Rendering a maximum of {} images per category<br/>'.format(
        options.max_images_per_category)
    
    html_output_string += '<br/>'
    
    html_output_string += ('Of {} total files:<br/><br/><div style="margin-left:15px;">{} common detections<br/>{} common non-detections<br/>{} A only<br/>{} B only<br/>{} class transitions</div><br/>'.format(
        len(filenames_to_compare),len(common_detections),
        len(common_non_detections),len(detections_a_only),
        len(detections_b_only),len(class_transitions)))
    
    html_output_string += 'Comparison pages:<br/><br/>\n'
    html_output_string += '<div style="margin-left:15px;">\n'
        
    comparison_path_relative = os.path.relpath(local_output_folder,options.output_folder)    
    for category in categories_to_image_pairs.keys():
        category_html_filename = os.path.join(comparison_path_relative,category + '.html')
        html_output_string += '<a href="{}">{}</a><br/>\n'.format(
            category_html_filename,category)
    
    html_output_string += '</div>\n'
    html_output_string += '</div>\n'
    
    pairwise_results = PairwiseBatchComparisonResults()
    
    pairwise_results.html_content = html_output_string
    pairwise_results.pairwise_options = pairwise_options
    pairwise_results.categories_to_image_pairs = categories_to_image_pairs
            
    return pairwise_results
        
# ...def _pairwise_compare_batch_results()


def compare_batch_results(options):
    """
    The main entry point for this module.  Runs one or more batch results comparisons, 
    writing results to an html page.  Most of the work is deferred to _pairwise_compare_batch_results().
    
    Args:
        options (BatchComparisonOptions): job options to use for this comparison task, including the
            list of specific pairswise comparisons to make (in the pairwise_options field)
            
    Returns:
        BatchComparisonResults: the results of this comparison task
    """
    
    assert options.output_folder is not None
    assert options.image_folder is not None
    assert options.pairwise_options is not None

    options = copy.deepcopy(options)
 
    if not isinstance(options.pairwise_options,list):
        options.pairwise_options = [options.pairwise_options]
    
    pairwise_options_list = options.pairwise_options
    n_comparisons = len(pairwise_options_list)
    
    options.pairwise_options = None
    
    html_content = ''
    all_pairwise_results = []
    
    # i_comparison = 0; pairwise_options = pairwise_options_list[i_comparison]
    for i_comparison,pairwise_options in enumerate(pairwise_options_list):
        print('Running comparison {} of {}'.format(i_comparison,n_comparisons))
        pairwise_results = \
            _pairwise_compare_batch_results(options,i_comparison,pairwise_options)
        html_content += pairwise_results.html_content
        all_pairwise_results.append(pairwise_results)

    html_output_string = main_page_header
    job_name_string = ''
    if len(options.job_name) > 0:
        job_name_string = ' for {}'.format(options.job_name)
    html_output_string += '<h2>Comparison of results{}</h2>\n'.format(
        job_name_string)
    html_output_string += html_content
    html_output_string += main_page_footer
    
    html_output_file = os.path.join(options.output_folder,'index.html')    
    with open(html_output_file,'w') as f:
        f.write(html_output_string) 
    
    results = BatchComparisonResults()
    results.html_output_file = html_output_file
    results.pairwise_results = all_pairwise_results
    return results


def n_way_comparison(filenames,options,detection_thresholds=None,rendering_thresholds=None):
    """
    Performs N pairwise comparisons for the list of results files in [filenames], by generating
    sets of pairwise options and calling compare_batch_results.
    
    Args:
        filenames (list): list of MD results filenames to compare
        options (BatchComparisonOptions): task options set in which pairwise_options is still 
            empty; that will get populated from [filenames]
        detection_thresholds (list, optional): list of detection thresholds with the same length
            as [filenames], or None to use sensible defaults
        rendering_thresholds (list, optional): list of rendering thresholds with the same length
            as [filenames], or None to use sensible defaults
    
    Returns:
        BatchComparisonResults: the results of this comparison task
    """
    
    if detection_thresholds is None:
        detection_thresholds = [0.15] * len(filenames)
    assert len(detection_thresholds) == len(filenames)

    if rendering_thresholds is not None:
        assert len(rendering_thresholds) == len(detection_thresholds)
    else:
        rendering_thresholds = [(x*0.6666) for x in detection_thresholds]

    # Choose all pairwise combinations of the files in [filenames]
    for i, j in itertools.combinations(list(range(0,len(filenames))),2):
            
        pairwise_options = PairwiseBatchComparisonOptions()
        
        pairwise_options.results_filename_a = filenames[i]
        pairwise_options.results_filename_b = filenames[j]
        
        pairwise_options.rendering_confidence_threshold_a = rendering_thresholds[i]
        pairwise_options.rendering_confidence_threshold_b = rendering_thresholds[j]
        
        pairwise_options.detection_thresholds_a = {'default':detection_thresholds[i]}
        pairwise_options.detection_thresholds_b = {'default':detection_thresholds[j]}
        
        options.pairwise_options.append(pairwise_options)

    return compare_batch_results(options)

# ...n_way_comparison()


#%% Interactive driver

if False:
        
    #%% Test two-way comparison
    
    options = BatchComparisonOptions()

    options.parallelize_rendering_with_threads = True
    
    options.job_name = 'BCT'
    options.output_folder = r'g:\temp\comparisons'
    options.image_folder = r'g:\camera_traps\camera_trap_images'
    options.max_images_per_category = 100
    options.sort_by_confidence = True
    
    options.pairwise_options = []

    results_base = os.path.expanduser('~/postprocessing/bellevue-camera-traps')
    filenames = [                
        os.path.join(results_base,r'bellevue-camera-traps-2023-12-05-v5a.0.0\combined_api_outputs\bellevue-camera-traps-2023-12-05-v5a.0.0_detections.json'),
        os.path.join(results_base,r'bellevue-camera-traps-2023-12-05-aug-v5a.0.0\combined_api_outputs\bellevue-camera-traps-2023-12-05-aug-v5a.0.0_detections.json')
        ]

    detection_thresholds = [0.15,0.15]
    rendering_thresholds = None
    
    results = n_way_comparison(filenames,options,detection_thresholds,rendering_thresholds=rendering_thresholds)
    
    from megadetector.utils.path_utils import open_file
    open_file(results.html_output_file)

    
    #%% Test three-way comparison
    
    options = BatchComparisonOptions()

    options.parallelize_rendering_with_threads = False

    options.job_name = 'KGA-test'
    options.output_folder = os.path.expanduser('~/tmp/md-comparison-test')
    options.image_folder = os.path.expanduser('~/data/KGA')

    options.pairwise_options = []

    filenames = [
        os.path.expanduser('~/data/KGA-4.json'),
        os.path.expanduser('~/data/KGA-5a.json'),
        os.path.expanduser('~/data/KGA-5b.json')
        ]

    detection_thresholds = [0.7,0.15,0.15]

    results = n_way_comparison(filenames,options,detection_thresholds,rendering_thresholds=None)
    
    from megadetector.utils.path_utils import open_file
    open_file(results.html_output_file)


#%% Command-line driver

"""
python compare_batch_results.py ~/tmp/comparison-test ~/data/KGA ~/data/KGA-5a.json ~/data/KGA-5b.json ~/data/KGA-4.json --detection_thresholds 0.15 0.15 0.7 --rendering_thresholds 0.1 0.1 0.6 --use_processes
"""

import sys,argparse,textwrap

def main():
    
    options = BatchComparisonOptions()
    
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=textwrap.dedent('''\
           Example:
        
           python compare_batch_results.py output_folder image_folder mdv5a.json mdv5b.json mdv4.json --detection_thresholds 0.15 0.15 0.7
           '''))
          
    parser.add_argument('output_folder', type=str, help='folder to which to write html results')

    parser.add_argument('image_folder', type=str, help='image source folder')

    parser.add_argument('results_files', nargs='*', type=str, help='list of .json files to be compared')

    parser.add_argument('--detection_thresholds', nargs='*', type=float,
                        help='list of detection thresholds, same length as the number of .json files, ' + \
                             'defaults to 0.15 for all files')
    
    parser.add_argument('--rendering_thresholds', nargs='*', type=float,
                        help='list of rendering thresholds, same length as the number of .json files, ' + \
                             'defaults to 0.10 for all files')
    
    parser.add_argument('--max_images_per_category', type=int, default=options.max_images_per_category,
                       help='number of images to sample for each agreement category (common detections, etc.)')
    
    parser.add_argument('--target_width', type=int, default=options.target_width, 
                        help='output image width, defaults to {}'.format(options.target_width))
    
    parser.add_argument('--use_processes', action='store_true', 
                        help='use processes rather than threads for parallelization')
    
    parser.add_argument('--open_results', action='store_true', 
                        help='open the output html file when done')
    
    parser.add_argument('--n_rendering_workers', type=int, default=options.n_rendering_workers,
                        help='number of workers for parallel rendering, defaults to {}'.format(
                            options.n_rendering_workers))
        
    if len(sys.argv[1:])==0:
        parser.print_help()
        parser.exit()
        
    args = parser.parse_args()
    
    print('Output folder:')
    print(args.output_folder)
    
    print('\nResults files:')
    print(args.results_files)
    
    print('\nDetection thresholds:')
    print(args.detection_thresholds)
    
    print('\nRendering thresholds:')
    print(args.rendering_thresholds)    
    
    # Convert to options objects
    options = BatchComparisonOptions()
    
    options.output_folder = args.output_folder
    options.image_folder = args.image_folder
    options.target_width = args.target_width
    options.n_rendering_workers = args.n_rendering_workers    
    options.max_images_per_category = args.max_images_per_category
    
    if args.use_processes:
        options.parallelize_rendering_with_threads = False
        
    results = n_way_comparison(args.results_files,options,args.detection_thresholds,args.rendering_thresholds)
    
    if args.open_results:
        path_utils.open_file(results.html_output_file)
        
    print('Wrote results to {}'.format(results.html_output_file))
    
# ...main()

    
if __name__ == '__main__':
    
    main()



"""

read_exif.py

Given a folder of images, reads relevant metadata (EXIF/IPTC/XMP) fields from all images, 
and writes them to  a .json or .csv file.  

This module can use either PIL (which can only reliably read EXIF data) or exiftool (which
can read everything).  The latter approach expects that exiftool is available on the system
path.  No attempt is made to be consistent in format across the two approaches.

"""

#%% Imports and constants

import os
import subprocess
import json
from datetime import date, datetime

from multiprocessing.pool import ThreadPool as ThreadPool
from multiprocessing.pool import Pool as Pool

from tqdm import tqdm
from PIL import Image, ExifTags

from megadetector.utils.path_utils import find_images, is_executable
from megadetector.utils.ct_utils import args_to_object
from megadetector.utils.ct_utils import image_file_to_camera_folder

debug_max_images = None


#%% Options

class ReadExifOptions:
    """
    Parameters controlling metadata extraction.
    """
    
    def __init__(self):
        
        #: Enable additional debug console output
        self.verbose = False
        
        #: If this is True and an output file is specified for read_exif_from_folder,
        #: and we encounter a serialization issue, we'll return the results but won't
        #: error.    
        self.allow_write_error = False
        
        #: Number of concurrent workers, set to <= 1 to disable parallelization
        self.n_workers = 1
        
        #: Should we use threads (vs. processes) for parallelization?
        #:
        #: Not relevant if n_workers is <= 1.
        self.use_threads = True
            
        #: "File" and "ExifTool" are tag types used by ExifTool to report data that 
        #: doesn't come from EXIF, rather from the file (e.g. file size).
        self.tag_types_to_ignore = set(['File','ExifTool'])
        
        #: Include/exclude specific tags (tags_to_include and tags_to_exclude are mutually incompatible)
        #:
        #: A useful set of tags one might want to limit queries for:
        #:
        #: options.tags_to_include = ['DateTime','Model','Make','ExifImageWidth','ExifImageHeight',
        #: 'DateTimeOriginal','Orientation']    
        self.tags_to_include = None
        
        #: Include/exclude specific tags (tags_to_include and tags_to_exclude are mutually incompatible)
        self.tags_to_exclude = None
        
        #: The command line to invoke if using exiftool, can be an absolute path to exiftool.exe, or
        #: can be just "exiftool", in which case it should be on your system path.
        self.exiftool_command_name = 'exiftool'
        
        #: How should we handle byte-formatted EXIF tags?
        #:
        #: 'convert_to_string': convert to a Python string
        #: 'delete': don't include at all
        #: 'raw': include as a byte string
        self.byte_handling = 'convert_to_string' # 'convert_to_string','delete','raw'
        
        #: Should we use exiftool or PIL?
        self.processing_library = 'pil' # 'exiftool','pil'
        

class ExifResultsToCCTOptions:
    """
    Options controlling the behavior of exif_results_to_cct() (which reformats the datetime information)
    extracted by read_exif_from_folder().
    """
    
    def __init__(self):
        
        #: Timestamps older than this are assumed to be junk; lots of cameras use a
        #: default time in 2000.
        self.min_valid_timestamp_year = 2001
    
        #: The EXIF tag from which to pull datetime information
        self.exif_datetime_tag = 'DateTimeOriginal'
    
        #: Function for extracting location information, should take a string
        #: and return a string.  Defaults to ct_utils.image_file_to_camera_folder.  If
        #: this is None, location is written as "unknown".
        self.filename_to_location_function = image_file_to_camera_folder
    

#%% Functions

def _get_exif_ifd(exif):
    """
    Read EXIF data from by finding the EXIF offset and reading tags directly
    
    https://github.com/python-pillow/Pillow/issues/5863
    """
    
    # Find the offset for all the EXIF information
    for key, value in ExifTags.TAGS.items():
        if value == "ExifOffset":
            break
    info = exif.get_ifd(key)
    return {
        ExifTags.TAGS.get(key, key): value
        for key, value in info.items()
    }


def read_pil_exif(im,options=None):
    """
    Read all the EXIF data we know how to read from an image, using PIL.  This is primarily
    an internal function; the main entry point for single-image EXIF information is 
    read_exif_tags_for_image().
    
    Args:
        im (str or PIL.Image.Image): image (as a filename or an Image object) from which 
            we should read EXIF data.
    
    Returns:
        dict: a dictionary mapping EXIF tag names to their values
    """
    
    if options is None:
        options = ReadExifOptions()
        
    image_name = '[image]'
    if isinstance(im,str):
        image_name = im
        im = Image.open(im)
        
    exif_tags = {}
    try:
        exif_info = im.getexif()
    except Exception:
        exif_info = None
        
    if exif_info is None:
        return exif_tags
    
    for k, v in exif_info.items():
        assert isinstance(k,str) or isinstance(k,int), \
            'Invalid EXIF key {}'.format(str(k))
        if k in ExifTags.TAGS:
            exif_tags[ExifTags.TAGS[k]] = str(v)
        else:
            # print('Warning: unrecognized EXIF tag: {}'.format(k))
            exif_tags[k] = str(v)
    
    exif_ifd_tags = _get_exif_ifd(exif_info)
    
    for k in exif_ifd_tags.keys():
        v = exif_ifd_tags[k]
        if k in exif_tags:
            if options.verbose:
                print('Warning: redundant EXIF values for {} in {}:\n{}\n{}'.format(
                    k,image_name,exif_tags[k],v))
        else:
            exif_tags[k] = v
    
    exif_tag_names = list(exif_tags.keys())
    
    # Type conversion and cleanup
    # 
    # Most quirky types will get serialized to string when we write to .json.
    for k in exif_tag_names:
        
        if isinstance(exif_tags[k],bytes):
            
            if options.byte_handling == 'delete':
                del exif_tags[k]
            elif options.byte_handling == 'raw':
                pass
            else:
                assert options.byte_handling == 'convert_to_string'
                exif_tags[k] = str(exif_tags[k])
        
        elif isinstance(exif_tags[k],str):
            
            exif_tags[k] = exif_tags[k].strip()
            
    return exif_tags

# ...read_pil_exif()


def format_datetime_as_exif_datetime_string(dt):
    """
    Returns a Python datetime object rendered using the standard EXIF datetime
    string format ('%Y:%m:%d %H:%M:%S')
    """
    
    return datetime.strftime(dt, '%Y:%m:%d %H:%M:%S')
    

def parse_exif_datetime_string(s,verbose=False):
    """"
    Exif datetimes are strings, but in a standard format:        
        
    %Y:%m:%d %H:%M:%S
    
    Parses one of those strings into a Python datetime object.
    
    Args:
        s (str): datetime string to parse, should be in standard EXIF datetime format
        verbose (bool, optional): enable additional debug output
    
    Returns:
        datetime: the datetime object created from [s]
    """
    
    dt = None
    try:
        dt = datetime.strptime(s, '%Y:%m:%d %H:%M:%S')
    except Exception:
        if verbose:
            print('Warning: could not parse datetime {}'.format(str(s)))
    return dt


def _filter_tags(tags,options):
    """
    Internal function used to include/exclude specific tags from the exif_tags
    dict.
    """
    
    if options is None:
        return tags
    if options.tags_to_include is None and options.tags_to_exclude is None:
        return tags
    if options.tags_to_include is not None:
        assert options.tags_to_exclude is None, "tags_to_include and tags_to_exclude are incompatible"
        tags_to_return = {}
        for tag_name in tags.keys():
            if tag_name in options.tags_to_include:
                tags_to_return[tag_name] = tags[tag_name]
        return tags_to_return
    if options.tags_to_exclude is not None:
        assert options.tags_to_include is None, "tags_to_include and tags_to_exclude are incompatible"
        tags_to_return = {}
        for tag_name in tags.keys():
            if tag_name not in options.tags_to_exclude:
                tags_to_return[tag_name] = tags[tag_name]
        return tags_to_return


def read_exif_tags_for_image(file_path,options=None):
    """
    Get relevant fields from EXIF data for an image
    
    Returns:
        dict: a dict with fields 'status' (str) and 'tags'. The exact format of 'tags' depends on 
        options (ReadExifOptions, optional): parameters controlling metadata extraction
        options.processing_library:
            
            - For exiftool, 'tags' is a list of lists, where each element is (type/tag/value)
            - For PIL, 'tags' is a dict (str:str)
    """
    
    if options is None:
        options = ReadExifOptions()
    
    result = {'status':'unknown','tags':[]}
    
    if options.processing_library == 'pil':
        
        try:
            exif_tags = read_pil_exif(file_path,options)

        except Exception as e:
            if options.verbose:
                print('Read failure for image {}: {}'.format(
                    file_path,str(e)))
            result['status'] = 'read_failure'
            result['error'] = str(e)
        
        if result['status'] == 'unknown':
            if exif_tags is None:            
                result['status'] = 'empty_read'
            else:
                result['status'] = 'success'
                result['tags'] = _filter_tags(exif_tags,options)
                            
        return result
        
    elif options.processing_library == 'exiftool':
        
        # -G means "Print group name for each tag", e.g. print:
        #
        # [File]          Bits Per Sample                 : 8
        #
        # ...instead of:
        #
        # Bits Per Sample                 : 8
        proc = subprocess.Popen([options.exiftool_command_name, '-G', file_path],
                                stdout=subprocess.PIPE, encoding='utf8')
        
        exif_lines = proc.stdout.readlines()    
        exif_lines = [s.strip() for s in exif_lines]
        if ( (exif_lines is None) or (len(exif_lines) == 0) or not \
            any([s.lower().startswith('[exif]') for s in exif_lines])):
            result['status'] = 'failure'
            return result
        
        # A list of three-element lists (type/tag/value)
        exif_tags = []
        
        # line_raw = exif_lines[0]
        for line_raw in exif_lines:
            
            # A typical line:
            #
            # [ExifTool]      ExifTool Version Number         : 12.13
            
            line = line_raw.strip()
            
            # Split on the first occurrence of ":"
            tokens = line.split(':',1)
            assert(len(tokens) == 2), 'EXIF tokenization failure ({} tokens, expected 2)'.format(
                len(tokens))
            
            field_value = tokens[1].strip()        
            
            field_name_type = tokens[0].strip()        
            field_name_type_tokens = field_name_type.split(None,1)
            assert len(field_name_type_tokens) == 2, 'EXIF tokenization failure'
            
            field_type = field_name_type_tokens[0].strip()
            assert field_type.startswith('[') and field_type.endswith(']'), \
                'Invalid EXIF field {}'.format(field_type)
            field_type = field_type[1:-1]
            
            if field_type in options.tag_types_to_ignore:
                if options.verbose:
                    print('Ignoring tag with type {}'.format(field_type))
                continue        
            
            field_name = field_name_type_tokens[1].strip()
            if options.tags_to_exclude is not None and field_name in options.tags_to_exclude:
                continue
            if options.tags_to_include is not None and field_name not in options.tags_to_include:
                continue
            tag = [field_type,field_name,field_value]
            
            exif_tags.append(tag)
            
        # ...for each output line
            
        result['status'] = 'success'
        result['tags'] = exif_tags
        return result
    
    else:
        
        raise ValueError('Unknown processing library {}'.format(
            options.processing_library))

    # ...which processing library are we using?
    
# ...read_exif_tags_for_image()


def _populate_exif_data(im, image_base, options=None):
    """
    Populate EXIF data into the 'exif_tags' field in the image object [im].
    
    im['file_name'] should be prepopulated, relative to image_base.
    
    Returns a modified version of [im], also modifies [im] in place.
    """
    
    if options is None:
        options = ReadExifOptions()

    fn = im['file_name']
    if options.verbose:
        print('Processing {}'.format(fn))
    
    try:
        
        file_path = os.path.join(image_base,fn)
        assert os.path.isfile(file_path), 'Could not find file {}'.format(file_path)
        result = read_exif_tags_for_image(file_path,options)
        if result['status'] == 'success':
            exif_tags = result['tags']            
            im['exif_tags'] = exif_tags
        else:
            im['exif_tags'] = None
            im['status'] = result['status']
            if 'error' in result:
                im['error'] = result['error']
            if options.verbose:
                print('Error reading EXIF data for {}'.format(file_path))
    
    except Exception as e:
        
        s = 'Error on {}: {}'.format(fn,str(e))
        print(s)
        im['error'] = s
        im['status'] = 'read failure'
        im['exif_tags'] = None
    
    return im

# ..._populate_exif_data()


def _create_image_objects(image_files,recursive=True):
    """
    Create empty image objects for every image in [image_files], which can be a 
    list of relative paths (which will get stored without processing, so the base 
    path doesn't matter here), or a folder name.
    
    Returns a list of dicts with field 'file_name' (a relative path).
    
    "recursive" is ignored if "image_files" is a list.
    """
    
    # Enumerate *relative* paths
    if isinstance(image_files,str):    
        print('Enumerating image files in {}'.format(image_files))
        assert os.path.isdir(image_files), 'Invalid image folder {}'.format(image_files)
        image_files = find_images(image_files,
                                  recursive=recursive,
                                  return_relative_paths=True,
                                  convert_slashes=True)
        
    images = []
    for fn in image_files:
        im = {}
        im['file_name'] = fn
        images.append(im)
    
    if debug_max_images is not None:
        print('Trimming input list to {} images'.format(debug_max_images))
        images = images[0:debug_max_images]
    
    return images


def _populate_exif_for_images(image_base,images,options=None):
    """
    Main worker loop: read EXIF data for each image object in [images] and 
    populate the image objects in place.
    
    'images' should be a list of dicts with the field 'file_name' containing
    a relative path (relative to 'image_base').    
    """
    
    if options is None:
        options = ReadExifOptions()

    if options.n_workers == 1:
      
        results = []
        for im in tqdm(images):
            results.append(_populate_exif_data(im,image_base,options))
        
    else:
        
        from functools import partial
        if options.use_threads:
            print('Starting parallel thread pool with {} workers'.format(options.n_workers))
            pool = ThreadPool(options.n_workers)
        else:
            print('Starting parallel process pool with {} workers'.format(options.n_workers))
            pool = Pool(options.n_workers)
    
        results = list(tqdm(pool.imap(partial(_populate_exif_data,image_base=image_base,
                                        options=options),images),total=len(images)))

    return results


def _write_exif_results(results,output_file):
    """
    Write EXIF information to [output_file].
    
    'results' is a list of dicts with fields 'exif_tags' and 'file_name'.

    Writes to .csv or .json depending on the extension of 'output_file'.         
    """
    
    if output_file.endswith('.json'):
        
        with open(output_file,'w') as f:
            json.dump(results,f,indent=1,default=str)
            
    elif output_file.endswith('.csv'):
        
        # Find all EXIF tags that exist in any image
        all_keys = set()
        for im in results:
            
            keys_this_image = set()
            exif_tags = im['exif_tags']
            file_name = im['file_name']
            for tag in exif_tags:
                tag_name = tag[1]
                assert tag_name not in keys_this_image, \
                    'Error: tag {} appears twice in image {}'.format(
                        tag_name,file_name)
                all_keys.add(tag_name)
                
            # ...for each tag in this image
            
        # ...for each image
        
        all_keys = sorted(list(all_keys))
        
        header = ['File Name']
        header.extend(all_keys)
        
        import csv
        with open(output_file,'w') as csvfile:
            
            writer = csv.writer(csvfile)
            
            # Write header
            writer.writerow(header)
            
            for im in results:
                
                row = [im['file_name']]
                kvp_this_image = {tag[1]:tag[2] for tag in im['exif_tags']}
                
                for i_key,key in enumerate(all_keys):
                    value = ''
                    if key in kvp_this_image:
                        value = kvp_this_image[key]
                    row.append(value)                                        
                # ...for each key that *might* be present in this image
                
                assert len(row) == len(header)
                
                writer.writerow(row)
                
            # ...for each image
            
        # ...with open()
    
    else:
        
        raise ValueError('Could not determine output type from file {}'.format(
            output_file))
        
    # ...if we're writing to .json/.csv
    
    print('Wrote results to {}'.format(output_file))

# ..._write_exif_results(...)


def read_exif_from_folder(input_folder,output_file=None,options=None,filenames=None,recursive=True):
    """
    Read EXIF data for a folder of images.
    
    Args:
        input_folder (str): folder to process; if this is None, [filenames] should be a list of absolute
            paths
        output_file (str, optional): .json file to which we should write results; if this is None, results
            are returned but not written to disk
        options (ReadExifOptions, optional): parameters controlling metadata extraction
        filenames (list, optional): allowlist of relative filenames (if [input_folder] is not None) or
            a list of absolute filenames (if [input_folder] is None)
        recursive (bool, optional): whether to recurse into [input_folder], not relevant if [input_folder]
            is None.
        verbose (bool, optional): enable additional debug output
            
    Returns:
        list: list of dicts, each of which contains EXIF information for one images.  Fields include at least:
            * 'file_name': the relative path to the image
            * 'exif_tags': a dict of EXIF tags whose exact format depends on [options.processing_library].
    """
    
    if options is None:
        options = ReadExifOptions()
    
    # Validate options
    if options.tags_to_include is not None:
        assert options.tags_to_exclude is None, "tags_to_include and tags_to_exclude are incompatible"
    if options.tags_to_exclude is not None:
        assert options.tags_to_include is None, "tags_to_include and tags_to_exclude are incompatible"    
    
    if input_folder is None:
        input_folder = ''
    if len(input_folder) > 0:
        assert os.path.isdir(input_folder), \
            '{} is not a valid folder'.format(input_folder)

    assert (len(input_folder) > 0) or (filenames is not None), \
        'Must specify either a folder or a list of files'
        
    if output_file is not None:    
        
        assert output_file.lower().endswith('.json') or output_file.lower().endswith('.csv'), \
            'I only know how to write results to .json or .csv'
            
        try:
            with open(output_file, 'a') as f:
                if not f.writable():
                    raise IOError('File not writable')
        except Exception:
            print('Could not write to file {}'.format(output_file))
            raise
        
    if options.processing_library == 'exif':
        assert is_executable(options.exiftool_command_name), 'exiftool not available'

    if filenames is None:
        images = _create_image_objects(input_folder,recursive=recursive)
    else:
        assert isinstance(filenames,list)
        images = _create_image_objects(filenames)
        
    results = _populate_exif_for_images(input_folder,images,options)
    
    if output_file is not None:
        try:
            _write_exif_results(results,output_file)
        except Exception as e:
            if not options.allow_write_error:
                raise
            else:
                print('Warning: error serializing EXIF data: {}'.format(str(e)))                
        
    return results

# ...read_exif_from_folder(...)


def exif_results_to_cct(exif_results,cct_output_file=None,options=None):
    """
    Given the EXIF results for a folder of images read via read_exif_from_folder,
    create a COCO Camera Traps .json file that has no annotations, but 
    attaches image filenames to locations and datetimes.
    
    Args:
        exif_results (str or list): the filename (or loaded list) containing the results
          from read_exif_from_folder
        cct_file (str,optional): the filename to which we should write COCO-Camera-Traps-formatted
          data
          
    Returns:
        dict: a COCO Camera Traps dict (with no annotations).
    """
    
    if options is None:
        options = ExifResultsToCCTOptions()
        
    if isinstance(exif_results,str):
        print('Reading EXIF results from {}'.format(exif_results))
        with open(exif_results,'r') as f:
            exif_results = json.load(f)
    else:
        assert isinstance(exif_results,list)
            
    now = datetime.now()

    image_info = []

    images_without_datetime = []
    images_with_invalid_datetime = []
    
    # exif_result = exif_results[0]
    for exif_result in tqdm(exif_results):
        
        im = {}
        
        # By default we assume that each leaf-node folder is a location
        if options.filename_to_location_function is None:
            im['location'] = 'unknown'
        else:
            im['location'] = options.filename_to_location_function(exif_result['file_name'])            
            
        im['file_name'] = exif_result['file_name']
        im['id'] = im['file_name']
        
        if ('exif_tags' not in exif_result) or (exif_result['exif_tags'] is None) or \
            (options.exif_datetime_tag not in exif_result['exif_tags']): 
            exif_dt = None
        else:
            exif_dt = exif_result['exif_tags'][options.exif_datetime_tag]
            exif_dt = parse_exif_datetime_string(exif_dt)
        if exif_dt is None:
            im['datetime'] = None
            images_without_datetime.append(im['file_name'])
        else:
            dt = exif_dt
            
            # An image from the future (or within the last 24 hours) is invalid
            if (now - dt).total_seconds() <= 1*24*60*60:
                print('Warning: datetime for {} is {}'.format(
                    im['file_name'],dt))
                im['datetime'] = None            
                images_with_invalid_datetime.append(im['file_name'])
            
            # An image from before the dawn of time is also invalid
            elif dt.year < options.min_valid_timestamp_year:
                print('Warning: datetime for {} is {}'.format(
                    im['file_name'],dt))
                im['datetime'] = None
                images_with_invalid_datetime.append(im['file_name'])
            
            else:
                im['datetime'] = dt

        image_info.append(im)
        
    # ...for each exif image result

    print('Parsed EXIF datetime information, unable to parse EXIF date from {} of {} images'.format(
        len(images_without_datetime),len(exif_results)))

    d = {}
    d['info'] = {}
    d['images'] = image_info
    d['annotations'] = []
    d['categories'] = []
    
    def json_serialize_datetime(obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        raise TypeError('Object {} (type {}) not serializable'.format(
            str(obj),type(obj)))
        
    if cct_output_file is not None:
        with open(cct_output_file,'w') as f:
            json.dump(d,f,indent=1,default=json_serialize_datetime)
    
    return d

# ...exif_results_to_cct(...)

    
#%% Interactive driver

if False:
    
    #%%
    
    input_folder = r'C:\temp\md-name-testing'
    output_file = None # r'C:\temp\md-name-testing\exif.json'
    options = ReadExifOptions()
    options.verbose = False
    options.n_workers = 10
    options.use_threads = False
    options.processing_library = 'pil'
    # options.processing_library = 'exiftool'
    options.tags_to_include = ['DateTime','Model','Make','ExifImageWidth','ExifImageHeight','DateTime','DateTimeOriginal','Orientation']
    # options.tags_to_exclude = ['MakerNote']
    
    results = read_exif_from_folder(input_folder,output_file,options)

    #%%
    
    with open(output_file,'r') as f:
        d = json.load(f)
        

#%% Command-line driver

import argparse
import sys

def main():

    options = ReadExifOptions()
    
    parser = argparse.ArgumentParser(description=('Read EXIF information from all images in' + \
                                                  ' a folder, and write the results to .csv or .json'))

    parser.add_argument('input_folder', type=str, 
                        help='Folder of images from which we should read EXIF information')
    parser.add_argument('output_file', type=str,
                        help='Output file (.json) to which we should write EXIF information')
    parser.add_argument('--n_workers', type=int, default=1,
                        help='Number of concurrent workers to use (defaults to 1)')
    parser.add_argument('--use_threads', action='store_true',
                        help='Use threads (instead of processes) for multitasking')
    parser.add_argument('--processing_library', type=str, default=options.processing_library,
                        help='Processing library (exif or pil)')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()    
    args_to_object(args, options)
    options.processing_library = options.processing_library.lower()
    
    read_exif_from_folder(args.input_folder,args.output_file,options)
    
if __name__ == '__main__':
    main()



"""

get_lila_annotation_counts.py

Generates a .json-formatted dictionary mapping each LILA dataset to all categories
that exist for that dataset, with counts for the number of occurrences of each category 
(the number of *annotations* for each category, not the number of *images*).

Also loads the taxonomy mapping file, to include scientific names for each category.

get_lila_image_counts.py counts the number of *images* for each category in each dataset.

"""

#%% Constants and imports

import json
import os

from megadetector.data_management.lila.lila_common import \
    read_lila_metadata, read_metadata_file_for_dataset, read_lila_taxonomy_mapping

# cloud provider to use for downloading images; options are 'gcp', 'azure', or 'aws'
preferred_cloud = 'gcp'

# array to fill for output
category_list = []

# We'll write images, metadata downloads, and temporary files here
lila_local_base = os.path.expanduser('~/lila')

output_dir = os.path.join(lila_local_base,'lila_categories_list')
os.makedirs(output_dir,exist_ok=True)

metadata_dir = os.path.join(lila_local_base,'metadata')
os.makedirs(metadata_dir,exist_ok=True)

output_file = os.path.join(output_dir,'lila_dataset_to_categories.json')


#%% Load category and taxonomy files

taxonomy_df = read_lila_taxonomy_mapping(metadata_dir)


#%% Map dataset names and category names to scientific names

ds_query_to_scientific_name = {}

unmapped_queries = set()

datasets_with_taxonomy_mapping = set()

# i_row = 1; row = taxonomy_df.iloc[i_row]; row
for i_row,row in taxonomy_df.iterrows():
    
    datasets_with_taxonomy_mapping.add(row['dataset_name'])
    
    ds_query = row['dataset_name'] + ':' + row['query']
    ds_query = ds_query.lower()
    
    if not isinstance(row['scientific_name'],str):
        unmapped_queries.add(ds_query)
        ds_query_to_scientific_name[ds_query] = 'unmapped'
        continue
        
    ds_query_to_scientific_name[ds_query] = row['scientific_name']
    
print('Loaded taxonomy mappings for {} datasets'.format(len(datasets_with_taxonomy_mapping)))
    

#%% Download and parse the metadata file

metadata_table = read_lila_metadata(metadata_dir)

print('Loaded metadata URLs for {} datasets'.format(len(metadata_table)))


#%% Download and extract metadata for each dataset

for ds_name in metadata_table.keys():    
    metadata_table[ds_name]['json_filename'] = read_metadata_file_for_dataset(ds_name=ds_name,
                                                                         metadata_dir=metadata_dir,
                                                                         metadata_table=metadata_table)


#%% Get category names and counts for each dataset

from collections import defaultdict

dataset_to_categories = {}

# ds_name = 'NACTI'
for ds_name in metadata_table.keys():
    
    taxonomy_mapping_available = (ds_name in datasets_with_taxonomy_mapping)
    
    if not taxonomy_mapping_available:
        print('Warning: taxonomy mapping not available for {}'.format(ds_name))
        
    print('Finding categories in {}'.format(ds_name))

    json_filename = metadata_table[ds_name]['json_filename']
    base_url = metadata_table[ds_name]['image_base_url_' + preferred_cloud]
    assert not base_url.endswith('/')
    
    # Open the metadata file    
    with open(json_filename, 'r') as f:
        data = json.load(f)
    
    # Collect list of categories and mappings to category name
    categories = data['categories']
    
    category_id_to_count = defaultdict(int)
    annotations = data['annotations']    
    
    # ann = annotations[0]
    for ann in annotations:
        category_id_to_count[ann['category_id']] = category_id_to_count[ann['category_id']] + 1
    
    # c = categories[0]
    for c in categories:
       count = category_id_to_count[c['id']] 
       if 'count' in c:
           assert 'bbox' in ds_name or c['count'] == count       
       c['count'] = count
       
       # Don't do taxonomy mapping for bbox data sets, which are sometimes just binary and are
       # always redundant with the class-level data sets.
       if 'bbox' in ds_name:
           c['scientific_name_from_taxonomy_mapping'] = None
       elif not taxonomy_mapping_available:
           c['scientific_name_from_taxonomy_mapping'] = None
       else:
           taxonomy_query_string = ds_name.lower().strip() + ':' + c['name'].lower()
           if taxonomy_query_string not in ds_query_to_scientific_name:
               print('No match for query string {}'.format(taxonomy_query_string))
               # As of right now, this is the only quirky case
               assert '#ref!' in taxonomy_query_string and 'wcs' in ds_name.lower()
               c['scientific_name_from_taxonomy_mapping'] = None
           else:
               sn = ds_query_to_scientific_name[taxonomy_query_string]
               assert sn is not None and len(sn) > 0
               c['scientific_name_from_taxonomy_mapping'] = sn
    
    dataset_to_categories[ds_name] = categories

# ...for each dataset


#%% Print the results

# ds_name = list(dataset_to_categories.keys())[0]
for ds_name in dataset_to_categories:
    
    print('\n** Category counts for {} **\n'.format(ds_name))
    
    categories = dataset_to_categories[ds_name]
    categories = sorted(categories, key=lambda x: x['count'], reverse=True)
    
    for c in categories:
        print('{} ({}): {}'.format(c['name'],c['scientific_name_from_taxonomy_mapping'],c['count']))
        
# ...for each dataset


#%% Save the results

with open(output_file, 'w') as f:
    json.dump(dataset_to_categories,f,indent=1)



"""

 idaho-camera-traps.py

 Prepare the Idaho Camera Traps dataset for release on LILA.

"""

#%% Imports and constants

import json
import os
import numpy as np
import dateutil 
import pandas as pd
import datetime
import shutil

from tqdm import tqdm
from bson import json_util

from collections import defaultdict

# Multi-threading for .csv file comparison and image existence validation
from multiprocessing.pool import Pool as Pool
from multiprocessing.pool import ThreadPool as ThreadPool
n_threads = 14
n_threads_file_copy = 20

input_base = r'i:\idfg-images'
output_base = r'h:\idaho-camera-traps'
output_image_base = r'j:\idaho-camera-traps-output'
assert os.path.isdir(input_base)
assert os.path.isdir(output_base)
assert os.path.isdir(output_image_base)

output_image_base_public = os.path.join(output_image_base,'public')
output_image_base_private = os.path.join(output_image_base,'private')

# We are going to map the original filenames/locations to obfuscated strings, but once
# we've done that, we will re-use the mappings every time we run this script. 
force_generate_mappings = False

# This is the file to which mappings get saved
id_mapping_file = os.path.join(output_base,'id_mapping.json')

# The maximum time (in seconds) between images within which two images are considered the 
# same sequence.
max_gap_within_sequence = 30

# This is a two-column file, where each line is [string in the original metadata],[category name we want to map it to]
category_mapping_file = os.path.join(output_base,'category_mapping.csv')

# The output file, using the original strings
output_json_original_strings = os.path.join(output_base,'idaho-camera-traps-original-strings.json')

# The output file, using obfuscated strings for everything but filenamed
output_json_remapped_ids = os.path.join(output_base,'idaho-camera-traps-remapped-ids.json')

# The output file, using obfuscated strings and obfuscated filenames
output_json = os.path.join(output_base,'idaho-camera-traps.json')

# One time only, I ran MegaDetector on the whole dataset...
megadetector_results_file = r'H:\idaho-camera-traps\idfg-2021-07-26idaho-camera-traps_detections.json'

# ...then set aside any images that *may* have contained humans that had not already been
# annotated as such.  Those went in this folder...
human_review_folder = os.path.join(output_base,'human_review')

# ...and the ones that *actually* had humans (identified via manual review) got
# copied to this folder...
human_review_selection_folder = os.path.join(output_base,'human_review_selections')

# ...which was enumerated to this text file, which is a manually-curated list of
# images that were flagged as human.
human_review_list = os.path.join(output_base,'human_flagged_images.txt')

# Unopinionated .json conversion of the .csv metadata
sequence_info_cache = os.path.join(output_base,'sequence_info.json')

valid_opstates = ['normal','maintenance','snow on lens','foggy lens','foggy weather',
                  'malfunction','misdirected','snow on lense','poop/slobber','sun','tilted','vegetation obstruction']
opstate_mappings = {'snow on lense':'snow on lens','poop/slobber':'lens obscured','maintenance':'human'}

survey_species_presence_columns = ['elkpresent','deerpresent','prongpresent']

presence_to_count_columns = {
    'otherpresent':['MooseAntlerless','MooseCalf','MooseOther','MooseBull','MooseUnkn',
                    'BlackBearAdult','BlackBearCub','LionAdult','LionKitten','WolfAdult',
                    'WolfPup','CattleCow','CattleCalf','other'],
    'elkpresent':['ElkSpike','ElkAntlerless','ElkCalf','ElkRaghorn','ElkMatBull','ElkUnkn','ElkPedNub'],
    'deerpresent':['MDbuck','MDantlerless','MDfawn','WTDbuck','WTDantlerless','WTDfawn','WTDunkn','MDunkn'],
    'prongpresent':['PronghornBuck','PronghornFawn','PHunkn']
    }

required_columns = ['File','Folder','Date','Time','otherpresent','other','otherwhat','opstate']
expected_presence_columns = ['elkpresent','deerpresent','prongpresent','humanpresent','otherpresent']
    
expected_count_columns = set()
for presence_column in presence_to_count_columns.keys():
    count_columns = presence_to_count_columns[presence_column]
    for count_column in count_columns:
        expected_count_columns.add(count_column)
        
def list_is_sorted(l):
    return all(l[i] <= l[i+1] for i in range(len(l)-1))


#%% List files (images + .csv)

def get_files():
    
    all_files_list = os.path.join(output_base,'all_files.json')
    force_file_enumeration = False
    
    if (os.path.isfile(all_files_list) and (not force_file_enumeration)):
        
        print('File list exists, bypassing enumeration')
        with open(all_files_list,'r') as f:
            all_files = json.load(f)
            
    else:
        
        from pathlib import Path
        all_files = []
        for path in Path(input_base).rglob('*.*'):
            path = str(path)
            path = os.path.relpath(path,input_base)
            all_files.append(path)
        with open(all_files_list,'w') as f:
            json.dump(all_files,f,indent=1)
    
    print('Enumerated {} files'.format(len(all_files)))
    
    image_files = [s for s in all_files if (s.lower().endswith('.jpg') or s.lower().endswith('.jpeg'))]
    csv_files = [s for s in all_files if (\
                                          (s.lower().endswith('.csv')) and \
                                          ('Backups' not in s) and \
                                          ('Metadata.csv' not in s) and \
                                          ('ExportedDataFiles' not in s) and \
                                          ('CSV Files' not in s)
                                              )]
    
    print('{} image files, {} .csv files'.format(len(image_files),len(csv_files)))
    
    # Ignore .csv files in folders with multiple .csv files
    
    # ...which would require some extra work to decipher.
    
    csv_files_to_ignore = []
    
    folder_to_csv_files = defaultdict(list)
    
    # fn = csv_files[0]
    for fn in csv_files:
        folder_name = os.path.dirname(fn)
        folder_to_csv_files[folder_name].append(fn)
    
    for folder_name in folder_to_csv_files.keys():
        if len(folder_to_csv_files[folder_name]) > 1:
            print('Multiple .csv files for {}:'.format(folder_name))
            for csv_file in folder_to_csv_files[folder_name]:
                print(csv_file)
                csv_files_to_ignore.append(csv_file)
            print('')
            
    n_csv_original = len(csv_files)
    csv_files = [s for s in csv_files if s not in csv_files_to_ignore]
    
    print('Processing {} of {} csv files'.format(len(csv_files),n_csv_original))

    return image_files,csv_files


#%% Parse each .csv file into sequences (function)

# csv_file = csv_files[-1]
def csv_to_sequences(csv_file):
    
    print('Processing {}'.format(csv_file))
    
    csv_file_absolute = os.path.join(input_base,csv_file)
    # os.startfile(csv_file_absolute)
    
    sequences = []
    # survey = csv_file.split('\\')[0]

    # Sample paths from which we need to derive locations:
    #
    # St.Joe_elk\AM99\Trip 1\100RECNX\TimelapseData.csv
    # Beaverhead_elk\AM34\Trip 1\100RECNX\TimelapseData.csv
    #
    # ClearCreek_mustelids\Winter2015-16\FS-001-P\FS-001-P.csv
    # ClearCreek_mustelids\Summer2015\FS-001\FS-001.csv
    # ClearCreek_mustelids\Summer2016\IDFG-016\IDFG-016.csv
    #
    # I:\idfg-images\ClearCreek_mustelids\Summer2016\IDFG-017b
    # I:\idfg-images\ClearCreek_mustelids\Summer2016\IDFG-017a
    if 'St.Joe_elk' in csv_file or 'Beaverhead_elk' in csv_file:
        location_name = '_'.join(csv_file.split('\\')[0:2]).replace(' ','')
    else:
        assert 'ClearCreek_mustelids' in csv_file
        tokens = csv_file.split('\\') 
        assert 'FS-' in tokens[2] or 'IDFG-' in tokens[2]
        location_name = '_'.join([tokens[0],tokens[2]]).replace('-P','')
        if location_name.endswith('017a') or location_name.endswith('017b'):
            location_name = location_name[:-1]
    
    # Load .csv file
    df = pd.read_csv(csv_file_absolute)
    df['datetime'] = None
    df['seq_id'] = None
    df['synthetic_frame_number'] = None
    
    # Validate the opstate column
    opstates = set(df['opstate'])
    for s in opstates:
        if isinstance(s,str):
            s = s.strip()
            if len(s) > 0:
                assert s in valid_opstates,'Invalid opstate: {}'.format(s)
    
    column_names = list(df.columns)
    
    for s in required_columns:
        assert s in column_names
    
    count_columns = [s for s in column_names if s in expected_count_columns]
    
    presence_columns = [s for s in column_names if s.endswith('present')]
    
    for s in presence_columns:
        if s not in expected_presence_columns:
            assert 'Unexpected presence column {} in {}'.format(s,csv_file)
    for s in expected_presence_columns:
        if s not in presence_columns:
            assert 'Missing presence column {} in {}'.format(s,csv_file)
    
    if False:
        for s in expected_count_columns:
            if s not in count_columns:
                print('Missing count column {} in {}'.format(s,csv_file))
        
    ## Create datetimes
    
    # print('Creating datetimes')
    
    # i_row = 0; row = df.iloc[i_row]
    for i_row,row in df.iterrows():
        
        date = row['Date']
        time = row['Time']
        datestring = date + ' ' + time
        dt = dateutil.parser.parse(datestring)
        assert dt.year >= 2015 and dt.year <= 2019
        df.loc[i_row,'datetime'] = dt
        
    # Make sure data are sorted chronologically
    #
    # In odd circumstances, they are not... so sort them first, but warn
    datetimes = list(df['datetime'])
    if not list_is_sorted(datetimes):
        print('Datetimes not sorted for {}'.format(csv_file))
    
    df = df.sort_values('datetime') 
    df.reset_index(drop=True, inplace=True)
    datetimes = list(df['datetime'])
    assert list_is_sorted(datetimes)

    # Debugging when I was trying to see what was up with the unsorted dates    
    if False:
        for i in range(0,len(datetimes)-1):
            dt = datetimes[i+1]
            prev_dt = datetimes[i]
            delta = dt - prev_dt
            assert delta >= datetime.timedelta(0)
    
    ## Parse into sequences    
    
    # print('Creating sequences')
    
    current_sequence_id = None
    next_frame_number = 0
    previous_datetime = None
        
    sequence_id_to_rows = defaultdict(list)
    
    # i_row = 0; row = df.iloc[i_row]    
    for i_row,row in df.iterrows():
        
        dt = row['datetime']
        assert dt is not None and isinstance(dt,datetime.datetime)
        
        # Start a new sequence if:
        #
        # * This image has no timestamp
        # * This image has a frame number of zero
        # * We have no previous image timestamp
        #
        if previous_datetime is None:
            delta = None
        else:
            delta = (dt - previous_datetime).total_seconds()
        
        # Start a new sequence if necessary
        if delta is None or delta > max_gap_within_sequence:
            next_frame_number = 0
            current_sequence_id = location_name + '_seq_' + str(dt) # str(uuid.uuid1())
            
        assert current_sequence_id is not None
        
        sequence_id_to_rows[current_sequence_id].append(i_row)
        df.loc[i_row,'seq_id'] = current_sequence_id
        df.loc[i_row,'synthetic_frame_number'] = next_frame_number
        next_frame_number = next_frame_number + 1
        previous_datetime = dt
        
    # ...for each row
    
    location_sequences = list(set(list(df['seq_id'])))
    location_sequences.sort()
    
    inconsistent_sequences = []
    
    
    ## Parse labels for each sequence
    
    # sequence_id = location_sequences[0]
    for sequence_id in location_sequences:
        
        sequence_row_indices = sequence_id_to_rows[sequence_id]
        assert len(sequence_row_indices) > 0
        
        # Row indices in a sequence should be adjacent
        if len(sequence_row_indices) > 1:
            d = np.diff(sequence_row_indices)
            assert(all(d==1))
        
        # sequence_df = df[df['seq_id']==sequence_id]
        sequence_df = df.iloc[sequence_row_indices]
        
        
        ## Determine what's present
        
        presence_columns_marked = []
        survey_species = []
        other_species = []
        
        # Be conservative; assume humans are present in all maintenance images
        opstates = set(sequence_df['opstate'])
        assert all([ ( (isinstance(s,float)) or (len(s.strip())== 0) or \
                      (s.strip() in valid_opstates)) for s in opstates]),\
            'Invalid optstate in: {}'.format(' | '.join(opstates))
        
        for presence_column in presence_columns:
                    
            presence_values = list(sequence_df[presence_column])
            
            # The presence columns are *almost* always identical for all images in a sequence        
            single_presence_value = (len(set(presence_values)) == 1)
            # assert single_presence_value
            if not single_presence_value:
                # print('Warning: presence value for {} is inconsistent for {}'.format(
                #   presence_column,sequence_id))
                inconsistent_sequences.append(sequence_id)                
            
            if any(presence_values):
                presence_columns_marked.append(presence_column)                
                
        # ...for each presence column
        
        # Tally up the standard (survey) species
        survey_species = [s.replace('present','') for s in presence_columns_marked if s != 'otherpresent']
        for opstate in opstates:
            if not isinstance(opstate,str):
                continue
            opstate = opstate.strip()
            if len(opstate) == 0:
                continue
            if opstate in opstate_mappings:
                    opstate = opstate_mappings[opstate]                
            if (opstate != 'normal') and (opstate not in survey_species):
                survey_species.append(opstate)
            
        # If no presence columns are marked, all counts should be zero
        if len(presence_columns_marked) == 0:
            
            # count_column = count_columns[0]
            for count_column in count_columns:
                
                values = list(set(list(sequence_df[count_column])))
                
                # Occasionally a count gets entered (correctly) without the presence column being marked
                # assert len(values) == 1 and values[0] == 0, 'Non-zero counts with no presence
                # columns marked for sequence {}'.format(sequence_id)
                if (not(len(values) == 1 and values[0] == 0)):
                    print('Warning: presence and counts are inconsistent for {}'.format(sequence_id))
                    
                    # Handle this by virtually checking the "right" box
                    for presence_column in presence_to_count_columns.keys():
                        count_columns_this_species = presence_to_count_columns[presence_column]
                        if count_column in count_columns_this_species:
                            if presence_column not in presence_columns_marked:
                                presence_columns_marked.append(presence_column)
                    
                    # Make sure we found a match
                    assert len(presence_columns_marked) > 0
                
        # Handle 'other' tags
        if 'otherpresent' in presence_columns_marked:
            
            sequence_otherwhats = set()
            sequence_comments = set()
            
            for i,r in sequence_df.iterrows():            
                otherwhat = r['otherwhat']
                if isinstance(otherwhat,str):
                    otherwhat = otherwhat.strip()
                    if len(otherwhat) > 0:
                        sequence_otherwhats.add(otherwhat)
                comment = r['comment']
                if isinstance(comment,str):
                    comment = comment.strip()
                    if len(comment) > 0:
                        sequence_comments.add(comment)
                
            freetext_species = []
            for s in sequence_otherwhats:
                freetext_species.append(s)
            for s in sequence_comments:
                freetext_species.append(s)
                
            counted_species = []
            
            otherpresent_columns = presence_to_count_columns['otherpresent']
            
            # column_name = otherpresent_columns[0]
            for column_name in otherpresent_columns:
            
                if column_name in sequence_df and column_name != 'other':
            
                    column_counts = list(sequence_df[column_name])
                    column_count_positive = any([c > 0 for c in column_counts])
                    
                    if column_count_positive:
                        # print('Found non-survey counted species column: {}'.format(column_name))
                        counted_species.append(column_name)
            
            # ...for each non-empty presence column
        
            # Very rarely, the "otherpresent" column is checked, but no more detail is available
            if not ( (len(freetext_species) > 0) or (len(counted_species) > 0) ):
                other_species.append('unknown')
                
            other_species += freetext_species
            other_species += counted_species
            
        # ...handling non-survey species
        
        all_species = other_species + survey_species
                 
        # Build the sequence data
        
        images = []
        # i_row = 0; row = sequence_df.iloc[i_row]
        for i_row,row in sequence_df.iterrows():
            im = {}
            # Only one folder used a single .csv file for two subfolders
            if ('RelativePath' in row) and (isinstance(row['RelativePath'],str)) \
                and (len(row['RelativePath'].strip()) > 0):
                assert 'IDFG-028' in location_name
                im['file_name'] = os.path.join(row['RelativePath'],row['File'])
            else:
                im['file_name'] = row['File']
            im['datetime'] = row['datetime']
            images.append(im)
            
        sequence = {}
        sequence['csv_source'] = csv_file
        sequence['sequence_id'] = sequence_id
        sequence['images'] = images
        sequence['species_present'] = all_species
        sequence['location'] = location_name
        
        sequences.append(sequence)
        
    # ...for each sequence

    return sequences

# ...def csv_to_sequences()


#%% Parse each .csv file into sequences (loop)

if __name__ == "__main__":
    
    #%%
    
    import multiprocessing
    multiprocessing.freeze_support()
    image_files,csv_files = get_files()
    
    #%%
    
    if n_threads == 1:
        
        # i_file = -1; csv_file = csv_files[i_file]
        sequences_by_file = []
        for i_file,csv_file in enumerate(csv_files):
            print('Processing file {} of {}'.format(i_file,len(csv_files)))
            sequences = csv_to_sequences(csv_file)
            sequences_by_file.append(sequences)
    
    else:
        
        pool = Pool(n_threads)
        sequences_by_file = list(pool.imap(csv_to_sequences,csv_files))

    
    #%% Save sequence data
        
    with open(sequence_info_cache,'w') as f:
        json.dump(sequences_by_file,f,indent=2,default=json_util.default)
        
    
    #%% Load sequence data
    
    if False:
        
        #%%
    
        with open(sequence_info_cache,'r') as f:
            sequences_by_file = json.load(f,object_hook=json_util.object_hook)
        
        
    #%% Validate file mapping (based on the existing enumeration)
    
    missing_images = []
    image_files_set = set(image_files)
    n_images_in_sequences = 0
    sequence_ids = set()
    
    # sequences = sequences_by_file[0]
    for i_sequences,sequences in enumerate(tqdm(sequences_by_file)):
        
        assert len(sequences) > 0
        csv_source = sequences[0]['csv_source']
        csv_file_absolute = os.path.join(input_base,csv_source)
        csv_folder = os.path.dirname(csv_file_absolute)
        assert os.path.isfile(csv_file_absolute)
        
        # sequence = sequences[0]
        for i_sequence,sequence in enumerate(sequences):
            
            assert sequence['csv_source'] == csv_source
            sequence_id = sequence['sequence_id']
            if sequence_id in sequence_ids:
                print('Warning: duplicate sequence for {}, creating new sequence'.format(sequence_id))
                sequence['sequence_id'] = sequence['sequence_id'] + '_' + str(i_sequences) + \
                    '_' + str(i_sequence)
                sequence_id = sequence['sequence_id']
                assert sequence_id not in sequence_ids
                
            sequence_ids.add(sequence_id)
            
            species_present = sequence['species_present']
            images = sequence['images']
            
            for im in images:
        
                n_images_in_sequences += 1
                image_file_relative = im['file_name']
                
                # Actually, one folder has relative paths
                # assert '\\' not in image_file_relative and '/' not in image_file_relative
                
                image_file_absolute = os.path.join(csv_folder,image_file_relative)
                image_file_container_relative = os.path.relpath(image_file_absolute,input_base)
                
                # os.startfile(csv_folder)
                # assert os.path.isfile(image_file_absolute)
                # found_file = os.path.isfile(image_file_absolute)
                found_file = image_file_container_relative in image_files_set
                if not found_file:
                    print('Warning: can\'t find image {}'.format(image_file_absolute))
                    missing_images.append(image_file_absolute)
                    
            # ...for each image
    
        # ...for each sequence
    
    # ...for each .csv file            
    
    print('{} of {} images missing ({} on disk)'.format(len(missing_images),n_images_in_sequences,
                                                        len(image_files)))
            
    
    #%% Load manual category mappings
    
    with open(category_mapping_file,'r') as f:
        category_mapping_lines = f.readlines()
        category_mapping_lines = [s.strip() for s in category_mapping_lines]

    category_mappings = {}
    for s in category_mapping_lines:
        tokens = s.split(',',1)
        category_name = tokens[0].strip()
        category_value = tokens[1].strip().replace('"','').replace(',','+')
        assert ',' not in category_name
        assert ',' not in category_value
        
        # The second column is blank when the first column already represents the category name
        if len(category_value) == 0:
            category_value = category_name
        category_mappings[category_name] = category_value
                
    
    #%% Convert to CCT .json (original strings)
            
    human_flagged_images = []
    with open(human_review_list,'r') as f:
        human_flagged_images = f.readlines()
        human_flagged_images = [s.strip().replace('/','\\') for s in human_flagged_images]
        human_flagged_images = set(human_flagged_images)
    print('Read {} human flagged images'.format(len(human_flagged_images)))
    
    annotations = []
    image_id_to_image = {}
    category_name_to_category = {}
    
    # Force the empty category to be ID 0
    empty_category_id = 0
    empty_category = {}
    empty_category['id'] = empty_category_id
    empty_category['name'] = 'empty'
    category_name_to_category['empty'] = empty_category
    
    human_category_id = 1
    human_category = {}
    human_category['id'] = human_category_id
    human_category['name'] = 'human'
    category_name_to_category['human'] = human_category
    
    next_category_id = 2
    
    annotation_ids = set()
    
    if False:
        target_folder = r'ClearCreek_mustelids\Summer2015\FS-035'
        for sequences in sequences_by_file:
            if target_folder in sequences[0]['csv_source']:
                break
            
    # For each .csv file...
    #
    # sequences = sequences_by_file[0]
    for sequences in tqdm(sequences_by_file):
            
        # For each sequence...
        #
        # sequence = sequences[0]
        for sequence in sequences:
        
            species_present = sequence['species_present']
            species_present = [s.lower().strip().replace(',',';') for s in species_present]
            
            sequence_images = sequence['images']
            location = sequence['location'].lower().strip()
            sequence_id = sequence['sequence_id']
            csv_source = sequence['csv_source']
            csv_folder_relative = os.path.dirname(csv_source)
            
            sequence_category_ids = set()
            
            # Find categories for this image                
            if len(species_present) == 0:
                
                sequence_category_ids.add(0)
                assert category_name_to_category['empty']['id'] == list(sequence_category_ids)[0]
                
            else:
                                
                # When 'unknown' is used in combination with another label, use that
                # label; the "unknown" here doesn't mean "another unknown species", it means
                # there is some other unknown property about the main species.
                if 'unknown' in species_present and len(species_present) > 1:
                    assert all([((s in category_mappings) or (s in valid_opstates) or \
                                 (s in opstate_mappings.values()))\
                                for s in species_present if s != 'unknown'])
                    species_present = [s for s in species_present if s != 'unknown']
                                        
                # category_name_string = species_present[0]
                for category_name_string in species_present:
                    
                    # This piece of text had a lot of complicated syntax in it, and it would have 
                    # been too complicated to handle in a general way
                    if 'coyotoes' in category_name_string:
                        # print('Ignoring category {}'.format(category_name_string))
                        continue
                    
                    if category_name_string not in category_mappings:
                        assert category_name_string in valid_opstates or \
                            category_name_string in opstate_mappings.values()                            
                    else:
                        category_name_string = category_mappings[category_name_string]
                    assert ',' not in category_name_string
                    
                    category_names = category_name_string.split('+')
                    assert len(category_names) <= 2
                    
                    # Don't process redundant labels
                    category_names = set(category_names)
                    
                    # category_name = category_names[0]
                    for category_name in category_names:
                    
                        if category_name == 'ignore':
                            continue
                    
                        category_name = category_name.replace('"','')                            
                                                                        
                        # If we've seen this category before...
                        if category_name in category_name_to_category:
                                
                            category = category_name_to_category[category_name]
                            category_id = category['id'] 
                          
                        # If this is a new category...
                        else:
                            
                            # print('Adding new category for {}'.format(category_name))
                            category_id = next_category_id
                            category = {}
                            category['id'] = category_id
                            category['name'] = category_name
                            category_name_to_category[category_name] = category
                            next_category_id += 1
                            
                        sequence_category_ids.add(category_id)
                        
                    # ...for each category (inner)
                    
                # ...for each category (outer)
                    
            # ...if we do/don't have species in this sequence
            
            # We should have at least one category assigned (which may be "empty" or "unknown")
            assert len(sequence_category_ids) > 0
               
            # assert len(sequence_category_ids) > 0
                        
            # Was any image in this sequence manually flagged as human?
            for i_image,im in enumerate(sequence_images):
                
                file_name_relative = os.path.join(csv_folder_relative,im['file_name'])
                if file_name_relative in human_flagged_images:
                    # print('Flagging sequence {} as human based on manual review'.format(sequence_id))
                    assert human_category_id not in sequence_category_ids
                    sequence_category_ids.add(human_category_id)
                    break   
                    
            # For each image in this sequence...
            # 
            # i_image = 0; im = images[i_image]
            for i_image,im in enumerate(sequence_images):
                
                image_id = sequence_id + '_' + im['file_name']
                assert image_id not in image_id_to_image
                
                output_im = {}
                output_im['id'] = image_id
                output_im['file_name'] = os.path.join(csv_folder_relative,im['file_name'])
                output_im['seq_id'] = sequence_id
                output_im['seq_num_frames'] = len(sequence)
                output_im['frame_num'] = i_image
                output_im['datetime'] = str(im['datetime'])
                output_im['location'] = location
                
                image_id_to_image[image_id] = output_im
                
                # Create annotations for this image
                for i_ann,category_id in enumerate(sequence_category_ids):
                    
                    ann = {}
                    ann['id'] = 'ann_' + image_id + '_' + str(i_ann)
                    assert ann['id'] not in annotation_ids
                    annotation_ids.add(ann['id'])
                    ann['image_id'] = image_id
                    ann['category_id'] = category_id
                    ann['sequence_level_annotation'] = True
                    annotations.append(ann)

            # ...for each image in this sequence

        # ...for each sequence
                    
    # ...for each .csv file
    
    images = list(image_id_to_image.values())
    categories = list(category_name_to_category.values())
    print('Loaded {} annotations in {} categories for {} images'.format(
        len(annotations),len(categories),len(images)))
    
    # Verify that all images have annotations
    image_id_to_annotations = defaultdict(list)
    
    # ann = ict_data['annotations'][0]
    
    # For debugging only
    categories_to_counts = defaultdict(int)
    for ann in tqdm(annotations):
        image_id_to_annotations[ann['image_id']].append(ann)
        categories_to_counts[ann['category_id']] = categories_to_counts[ann['category_id']] + 1
        
    for im in tqdm(images):        
        image_annotations = image_id_to_annotations[im['id']]
        assert len(image_annotations) > 0
        
    
    #%% Create output (original strings)
    
    info = {}
    info['contributor'] = 'Idaho Department of Fish and Game'
    info['description'] = 'Idaho Camera traps'    
    info['version'] = '2021.07.19'
    
    output_data = {}
    output_data['images'] = images
    output_data['annotations'] = annotations
    output_data['categories'] = categories
    output_data['info'] = info
            
    with open(output_json_original_strings,'w') as f:
        json.dump(output_data,f,indent=1)
        
    
    #%% Validate .json file

    from megadetector.data_management.databases import integrity_check_json_db

    options = integrity_check_json_db.IntegrityCheckOptions()
    options.baseDir = input_base
    options.bCheckImageSizes = False
    options.bCheckImageExistence = False
    options.bFindUnusedImages = False

    _, _, _ = integrity_check_json_db.integrity_check_json_db(output_json_original_strings, options)
    
    
    #%% Preview labels
            
    from megadetector.visualization import visualize_db
    
    viz_options = visualize_db.DbVizOptions()
    viz_options.num_to_visualize = 1000
    viz_options.trim_to_images_with_bboxes = False
    viz_options.add_search_links = False
    viz_options.sort_by_filename = False
    viz_options.parallelize_rendering = True
    viz_options.include_filename_links = True
    
    viz_options.classes_to_exclude = ['empty','deer','elk']
    html_output_file, _ = visualize_db.visualize_db(db_path=output_json_original_strings,
                                                             output_dir=os.path.join(
                                                             output_base,'preview'),
                                                             image_base_dir=input_base,
                                                             options=viz_options)
    os.startfile(html_output_file)
    
    
    #%% Look for humans that were found by MegaDetector that haven't already been identified as human
    
    # This whole step only needed to get run once
    
    if False:
        
        pass
    
        #%%
        
        human_confidence_threshold = 0.5
        
        # Load MD results
        with open(megadetector_results_file,'r') as f:
            md_results = json.load(f)
            
        # Get a list of filenames that MD tagged as human
        
        human_md_categories =\
            [category_id for category_id in md_results['detection_categories'] if \
             ((md_results['detection_categories'][category_id] == 'person') or \
              (md_results['detection_categories'][category_id] == 'vehicle'))]
        assert len(human_md_categories) == 2
            
        # im = md_results['images'][0]
        md_human_images = set()
        
        for im in md_results['images']:
            if 'detections' not in im:
                continue
            if im['max_detection_conf'] < human_confidence_threshold:
                  continue
            for detection in im['detections']:
                if detection['category'] not in human_md_categories:
                    continue
                elif detection['conf'] < human_confidence_threshold:
                    continue
                else:
                    md_human_images.add(im['file'])
                    break
            
            # ...for each detection
            
        # ...for each image
        
        print('MD found {} potential human images (of {})'.format(
            len(md_human_images),len(md_results['images'])))    
    
        # Map images to annotations in ICT
        
        with open(output_json_original_strings,'r') as f:
            ict_data = json.load(f)
        
        category_id_to_name = {c['id']:c['name'] for c in categories}
        
        image_id_to_annotations = defaultdict(list)
        
        # ann = ict_data['annotations'][0]
        for ann in tqdm(ict_data['annotations']):
            image_id_to_annotations[ann['image_id']].append(ann)
            
        human_ict_categories = ['human']
        manual_human_images = set()
        
        # For every image
        # im = ict_data['images'][0]
        for im in tqdm(ict_data['images']):
                
            # Does this image already have a human annotation?
            manual_human = False
            
            annotations = image_id_to_annotations[im['id']]
            assert len(annotations) > 0
            
            for ann in annotations:
                category_name = category_id_to_name[ann['category_id']]
                if category_name in human_ict_categories:        
                    manual_human_images.add(im['file_name'].replace('\\','/'))
            
            # ...for each annotation
            
        # ...for each image
        
        print('{} images identified as human in source metadata'.format(len(manual_human_images)))
            
        missing_human_images = []
        
        for fn in md_human_images:
            if fn not in manual_human_images:
                missing_human_images.append(fn)
        
        print('{} potentially untagged human images'.format(len(missing_human_images)))
    
    
        #%% Copy images for review to a new folder
                
        os.makedirs(human_review_folder,exist_ok=True)
        missing_human_images.sort()
        
        # fn = missing_human_images[0]
        for i_image,fn in enumerate(tqdm(missing_human_images)):
            input_fn_absolute = os.path.join(input_base,fn).replace('\\','/')
            assert os.path.isfile(input_fn_absolute)
            output_path = os.path.join(human_review_folder,str(i_image).zfill(4) + '_' + fn.replace('/','~'))
            shutil.copyfile(input_fn_absolute,output_path)
        
        
        #%% Manual step...
        
        # Copy any images from that list that have humans in them to...
        human_review_selection_folder = r'H:\idaho-camera-traps\human_review_selections'
        assert os.path.isdir(human_review_selection_folder)
        
        
        #%% Create a list of the images we just manually flagged
        
        human_tagged_filenames = os.listdir(human_review_selection_folder)
        human_tagged_relative_paths = []
        # fn = human_tagged_filenames[0]
        for fn in human_tagged_filenames:
            
            # E.g. '0000_Beaverhead_elk~AM174~Trip 1~100RECNX~IMG_1397.JPG'
            relative_path = fn[5:].replace('~','/')
            human_tagged_relative_paths.append(relative_path)
        
        with open(human_review_list,'w') as f:
            for s in human_tagged_relative_paths:
                f.write(s + '\n')
            

    #%% Translate location, image, sequence IDs
    
    # Load mappings if available        
    if (not force_generate_mappings) and (os.path.isfile(id_mapping_file)):
        
        print('Loading ID mappings from {}'.format(id_mapping_file))
        
        with open(id_mapping_file,'r') as f:
            mappings = json.load(f)
    
        image_id_mappings = mappings['image_id_mappings']
        annotation_id_mappings = mappings['annotation_id_mappings']
        location_id_mappings = mappings['location_id_mappings']
        sequence_id_mappings = mappings['sequence_id_mappings']
        
    else:
        
        # Generate mappings
        mappings = {}
        
        next_location_id = 0
        location_id_string_to_n_sequences = defaultdict(int)
        location_id_string_to_n_images = defaultdict(int)
        
        image_id_mappings = {}
        annotation_id_mappings = {}
        location_id_mappings = {}
        sequence_id_mappings = {}
        
        for im in tqdm(images):
                            
            # If we've seen this location before...
            if im['location'] in location_id_mappings:
                location_id = location_id_mappings[im['location']]
            else:
                # Otherwise assign a string-formatted int as the ID
                location_id = str(next_location_id)
                
                location_id_mappings[im['location']] = location_id
                next_location_id += 1
            
            # If we've seen this sequence before...
            if im['seq_id'] in sequence_id_mappings:
                sequence_id = sequence_id_mappings[im['seq_id']]
            else:
                # Otherwise assign a string-formatted int as the ID
                n_sequences_this_location = location_id_string_to_n_sequences[location_id]
                sequence_id = 'loc_{}_seq_{}'.format(
                    location_id.zfill(4),str(n_sequences_this_location).zfill(6))
                sequence_id_mappings[im['seq_id']] = sequence_id
                
                n_sequences_this_location += 1
                location_id_string_to_n_sequences[location_id] = n_sequences_this_location
                
            assert im['id'] not in image_id_mappings
            
            # Assign an image ID
            
            n_images_this_location = location_id_string_to_n_images[location_id]                
            image_id_mappings[im['id']] = 'loc_{}_im_{}'.format(
                location_id.zfill(4),str(n_images_this_location).zfill(6))
            
            n_images_this_location += 1
            location_id_string_to_n_images[location_id] = n_images_this_location
        
        # ...for each image
        
        # Assign annotation mappings
        for i_ann,ann in enumerate(tqdm(annotations)):
            assert ann['image_id'] in image_id_mappings
            assert ann['id'] not in annotation_id_mappings
            annotation_id_mappings[ann['id']] = 'ann_{}'.format(str(i_ann).zfill(8))
            
        mappings['image_id_mappings'] = image_id_mappings
        mappings['annotation_id_mappings'] = annotation_id_mappings
        mappings['location_id_mappings'] = location_id_mappings
        mappings['sequence_id_mappings'] = sequence_id_mappings
                    
        # Save mappings
        with open(id_mapping_file,'w') as f:
            json.dump(mappings,f,indent=2)
            
        print('Saved ID mappings to {}'.format(id_mapping_file))
       
        # Back this file up, lest we should accidentally re-run this script
        # with force_generate_mappings = True and overwrite the mappings we used.
        datestr = str(datetime.datetime.now()).replace(':','-')
        backup_file = id_mapping_file.replace('.json','_' + datestr + '.json')
        shutil.copyfile(id_mapping_file,backup_file)
        
    # ...if we are/aren't re-generating mappings
    
    
    #%% Apply mappings
    
    for im in images:
        im['id'] = image_id_mappings[im['id']]
        im['seq_id'] = sequence_id_mappings[im['seq_id']]
        im['location'] = location_id_mappings[im['location']]
    for ann in annotations:
        ann['id'] = annotation_id_mappings[ann['id']]
        ann['image_id'] = image_id_mappings[ann['image_id']]
    
    print('Applied mappings')
    
    
    #%% Write new dictionaries (modified strings, original files)
    
    output_data = {}
    output_data['images'] = images
    output_data['annotations'] = annotations
    output_data['categories'] = categories
    output_data['info'] = info
    
    with open(output_json_remapped_ids,'w') as f:
        json.dump(output_data,f,indent=2)
        
    
    #%% Validate .json file (modified strings, original files)

    from megadetector.data_management.databases import integrity_check_json_db

    options = integrity_check_json_db.IntegrityCheckOptions()
    options.baseDir = input_base
    options.bCheckImageSizes = False
    options.bCheckImageExistence = False
    options.bFindUnusedImages = False

    _, _, _ = integrity_check_json_db.integrity_check_json_db(output_json_remapped_ids, options)
    
    
    #%% Preview labels (original files)
            
    from megadetector.visualization import visualize_db
    
    viz_options = visualize_db.DbVizOptions()
    viz_options.num_to_visualize = 1000
    viz_options.trim_to_images_with_bboxes = False
    viz_options.add_search_links = False
    viz_options.sort_by_filename = False
    viz_options.parallelize_rendering = True
    viz_options.include_filename_links = True
    
    # viz_options.classes_to_exclude = ['empty','deer','elk']
    # viz_options.classes_to_include = ['bobcat']
    viz_options.classes_to_include = [viz_options.multiple_categories_tag] 
    
    html_output_file, _ = visualize_db.visualize_db(db_path=output_json_remapped_ids,
                                                             output_dir=os.path.join(
                                                             output_base,'preview'),
                                                             image_base_dir=input_base,
                                                             options=viz_options)
    os.startfile(html_output_file)


    #%% Copy images to final output folder (prep)
    
    force_copy = False
    
    with open(output_json_remapped_ids,'r') as f:
        d = json.load(f)
        
    images = d['images']
    
    private_categories = ['human','domestic dog','vehicle']
    
    private_image_ids = set()

    category_id_to_name = {c['id']:c['name'] for c in d['categories']}
    
    # ann = d['annotations'][0]
    for ann in d['annotations']:
        category_name = category_id_to_name[ann['category_id']]
        if category_name in private_categories:
            private_image_ids.add(ann['image_id'])
    
    print('Moving {} of {} images to the private folder'.format(len(private_image_ids),len(images)))
    
    def process_image(im):
        
        input_relative_path = im['file_name']
        input_absolute_path = os.path.join(input_base,input_relative_path)
        
        if not os.path.isfile(input_absolute_path):
            print('Warning: file {} is not available'.format(input_absolute_path))
            return
        
        location = im['location']
        image_id = im['id']
        
        location_folder = 'loc_' + location.zfill(4)
        assert location_folder in image_id
        
        output_relative_path =  location_folder + '/' + image_id + '.jpg'
        
        # Is this a public or private image?
        private_image = (image_id in private_image_ids)
                
        # Generate absolute path
        if private_image:
            output_absolute_path = os.path.join(output_image_base_private,output_relative_path)
        else:
            output_absolute_path = os.path.join(output_image_base_public,output_relative_path)
                
        # Copy to output
        output_dir = os.path.dirname(output_absolute_path)
        os.makedirs(output_dir,exist_ok=True)
        
        if force_copy or (not os.path.isfile(output_absolute_path)):
            shutil.copyfile(input_absolute_path,output_absolute_path)
    
        # Update the filename reference
        im['file_name'] = output_relative_path

    # ...def process_image(im)
    
    
    #%% Copy images to final output folder (execution)
        
    # For each image
    if n_threads_file_copy == 1:
        # im = images[0]
        for im in tqdm(images):    
            process_image(im)
    else:
        pool = ThreadPool(n_threads_file_copy)
        pool.map(process_image,images)

    print('Finished copying, writing .json output')
    
    # Write output .json
    with open(output_json,'w') as f:
        json.dump(d,f,indent=1)
    
    
    #%% Make sure the right number of images got there
    
    from pathlib import Path
    all_output_files = []
    all_output_files_list = os.path.join(output_base,'all_output_files.json')
    
    for path in Path(output_image_base).rglob('*.*'):
        path = str(path)
        path = os.path.relpath(path,output_image_base)
        all_output_files.append(path)
    with open(all_output_files_list,'w') as f:
        json.dump(all_output_files,f,indent=1)

    print('Enumerated {} output files (of {} images)'.format(len(all_output_files),len(images)))


    #%% Validate .json file (final filenames)

    from megadetector.data_management.databases import integrity_check_json_db

    options = integrity_check_json_db.IntegrityCheckOptions()
    options.baseDir = input_base
    options.bCheckImageSizes = False
    options.bCheckImageExistence = False
    options.bFindUnusedImages = False

    _, _, _ = integrity_check_json_db.integrity_check_json_db(output_json, options)
    
    
    #%% Preview labels (final filenames)
    
    from megadetector.visualization import visualize_db
    
    viz_options = visualize_db.DbVizOptions()
    viz_options.num_to_visualize = 1500
    viz_options.trim_to_images_with_bboxes = False
    viz_options.add_search_links = False
    viz_options.sort_by_filename = False
    viz_options.parallelize_rendering = True
    viz_options.include_filename_links = True
    
    # viz_options.classes_to_exclude = ['empty','deer','elk']
    viz_options.classes_to_include = ['bear','mountain lion']
    # viz_options.classes_to_include = ['horse']
    # viz_options.classes_to_include = [viz_options.multiple_categories_tag] 
    # viz_options.classes_to_include = ['human','vehicle','domestic dog'] 
    
    html_output_file, _ = visualize_db.visualize_db(db_path=output_json,
                                                             output_dir=os.path.join(
                                                             output_base,'final-preview-01'),
                                                             image_base_dir=output_image_base_public,
                                                             options=viz_options)
    os.startfile(html_output_file)


    #%% Create zipfiles
    
    #%% List public files
    
    from pathlib import Path
    all_public_output_files = []
    all_public_output_files_list = os.path.join(output_base,'all_public_output_files.json')
    
    if not os.path.isfile(all_public_output_files_list):
        for path in Path(output_image_base_public).rglob('*.*'):
            path = str(path)
            path = os.path.relpath(path,output_image_base)
            all_public_output_files.append(path)
        with open(all_public_output_files_list,'w') as f:
            json.dump(all_public_output_files,f,indent=1)
    else:
        with open(all_public_output_files_list,'r') as f:
            all_public_output_files = json.load(f)
            
    print('Enumerated {} public output files'.format(len(all_public_output_files)))


    #%% Find the size of each file
    
    filename_to_size = {}
    
    all_public_output_sizes_list = os.path.join(output_base,'all_public_output_sizes.json')
    
    if not os.path.isfile(all_public_output_sizes_list):
        # fn = all_public_output_files[0]
        for fn in tqdm(all_public_output_files):
            p = os.path.join(output_image_base,fn)
            assert os.path.isfile(p)
            filename_to_size[fn] = os.path.getsize(p)
            
        with open(all_public_output_sizes_list,'w') as f:
            json.dump(filename_to_size,f,indent=1)
    else:
        with open(all_public_output_sizes_list,'r') as f:
            filename_to_size = json.load(f)
                
    assert len(filename_to_size) == len(all_public_output_files)
    
    
    #%% Split into chunks of approximately-equal size
    
    import humanfriendly
    total_size = sum(filename_to_size.values())
    print('{} in {} files'.format(humanfriendly.format_size(total_size),len(all_public_output_files)))
    
    bytes_per_part = 320e9
    
    file_lists = []
        
    current_file_list = []
    n_bytes_current_file_list = 0
    
    for fn in all_public_output_files:
        size = filename_to_size[fn]
        current_file_list.append(fn)
        n_bytes_current_file_list += size
        if n_bytes_current_file_list > bytes_per_part:
            file_lists.append(current_file_list)
            current_file_list = []
            n_bytes_current_file_list = 0
    # ...for each file
    
    file_lists.append(current_file_list)
            
    assert sum([len(l) for l in file_lists]) == len(all_public_output_files)
        
    print('List sizes:')
    for l in file_lists:
        print(len(l))
    
    
    #%% Create a zipfile for each chunk
    
    from zipfile import ZipFile
    import zipfile
    import os
    
    def create_zipfile(i_file_list):
        
        file_list = file_lists[i_file_list]
        zipfile_name = os.path.join('k:\\idaho-camera-traps-images.part_{}.zip'.format(i_file_list))
        
        print('Processing archive {} to file {}'.format(i_file_list,zipfile_name))
        
        with ZipFile(zipfile_name, 'w') as zipObj:
            
            for filename_relative in file_list:
           
                assert filename_relative.startswith('public')
                filename_absolute = os.path.join(output_image_base,filename_relative)
                zipObj.write(filename_absolute.replace('\\','/'), 
                             filename_relative, compress_type=zipfile.ZIP_STORED)
                
            # ...for each filename
            
        # with ZipFile()
        
    # ...def create_zipfile()
    
    # i_file_list = 0; file_list = file_lists[i_file_list]
    
    n_zip_threads = 1 # len(file_lists)
    if n_zip_threads == 1:
        for i_file_list in range(0,len(file_lists)):
            create_zipfile(i_file_list)
    else:
        pool = ThreadPool(n_zip_threads)
        indices = list(range(0,len(file_lists)))
        pool.map(create_zipfile,indices)

# ....if __name__ == "__main__"


"""

 pc_to_json.py

 Convert a particular collection of .csv files from Parks Canada to CCT format.

"""

#%% Constants and environment

import pandas as pd
import uuid
import json
import time

import numpy as np
from tqdm import tqdm

import humanfriendly
import os
import PIL

from megadetector.data_management.databases import integrity_check_json_db
from megadetector.data_management.cct_json_utils import IndexedJsonDb
from megadetector.data_management import cct_json_to_filename_json
from megadetector.visualization import visualize_db
from megadetector.utils import path_utils

input_base = r"g:\20190715"
output_file = r"D:\wildlife_data\parks_canada\pc_20190715.json"
preview_base = r"D:\wildlife_data\parks_canada\preview"

filename_replacements = {}
category_mappings = {'':'unlabeled'}

csv_prefix = 'ImageData_Microsoft___'

expected_columns = 'Location,DateImage,TimeImage,Species,Total,Horses,DogsOnLeash,DogsOffLeash,AdultFemale,AdultMale,AdultUnknown,Subadult,YLY,YOY,ImageName'.split(',')
columns_to_copy = {'Total':'count','Horses':'horses','DogsOnLeash':'dogsonleash','DogsOffLeash':'dogsoffleash',
                   'AdultFemale':'adultfemale','AdultMale':'adultmale','AdultUnknown':'adultunknown',
                   'Subadult':'subadult','YLY':'yearling','YOY':'youngofyear'}

retrieve_image_sizes = False

max_num_csvs = -1

db_sampling_scheme = 'preview' # 'labeled','all'
n_unlabeled_to_sample = -1
cap_unlabeled_to_labeled = True


#%% Read and concatenate source data

# List files
input_files = os.listdir(input_base)

# List of dataframes, one per .csv file; we'll concatenate later
all_input_metadata = []

# i_file = 87; fn = input_files[i_file]
for i_file,fn in enumerate(input_files):

    if max_num_csvs > 0 and len(all_input_metadata) >= max_num_csvs:
        break
    
    if not fn.endswith('.csv'):
        continue
    if not fn.startswith(csv_prefix):
        continue
    dirname = fn.replace(csv_prefix,'').replace('.csv','')
    dirfullpath = os.path.join(input_base,dirname)
    if not os.path.isdir(dirfullpath):
        dirname = fn.replace(csv_prefix,'').replace('.csv','').replace('  ',' ')
        dirfullpath = os.path.join(input_base,dirname)    
    assert(os.path.isdir(dirfullpath))
    
    metadata_fullpath = os.path.join(input_base,fn)
    print('Reading {}'.format(metadata_fullpath))
    df = pd.read_csv(metadata_fullpath)
    assert list(df.columns) == expected_columns
    df['DirName'] = dirname
    all_input_metadata.append(df)

# Concatenate into a giant data frame
input_metadata = pd.concat(all_input_metadata)

print('Read {} rows total'.format(len(input_metadata)))


#%% List files

print('Listing images...')
image_full_paths = path_utils.find_images(input_base,bRecursive=True)
print('Finished listing {} images'.format(len(image_full_paths)))

image_relative_paths = []
for s in image_full_paths:
    image_relative_paths.append(os.path.relpath(s,input_base))
image_relative_paths = set(image_relative_paths)

image_relative_paths_lower = set()
for s in image_relative_paths:
    image_relative_paths_lower.add(s.lower())
    
    
#%% Main loop over labels (prep)

start_time = time.time()

relative_path_to_image = {}

images = []
annotations = []
category_name_to_category = {}
missing_files = []

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
category_name_to_category['empty'] = empty_category
next_category_id = 1

labeled_images = []
unlabeled_images = []


#%% Main loop over labels (loop)

# iRow = 0; row = input_metadata.iloc[iRow]
for iRow,row in tqdm(input_metadata.iterrows(),total=len(input_metadata)):
    
    # ImageID,FileName,FilePath,SpeciesID,CommonName
    image_id = str(uuid.uuid1())
    relative_path = os.path.normpath(row['ImageName'])
    
    if relative_path not in image_relative_paths:
        if relative_path.lower() in image_relative_paths_lower:
            print('Warning: lower-case version of {} in path list'.format(relative_path))
        else:
            missing_files.append(relative_path)
            continue
        
    full_path = os.path.join(input_base,relative_path)

    # assert os.path.isfile(full_path)    
 
    if relative_path in relative_path_to_image:
        
        im = relative_path_to_image[relative_path]

    else:
        
        im = {}
        im['id'] = image_id
        im['file_name'] = relative_path
        im['seq_id'] = '-1'
        im['location'] = row['Location']
        im['datetime'] = row['DateImage'] + ' ' + row['TimeImage']        
        
        images.append(im)
        relative_path_to_image[relative_path] = im
        
        if retrieve_image_sizes:
        
            # Retrieve image width and height
            pil_im = PIL.Image.open(full_path)
            width, height = pil_im.size
            im['width'] = width
            im['height'] = height

    species = row['Species']
    if isinstance(species,float):
        assert np.isnan(species)
        species = 'unlabeled'
        
    category_name = species.lower().strip()
    if category_name in category_mappings:
        category_name = category_mappings[category_name]
        
    if category_name not in category_name_to_category:
        category = {}
        category['name'] = category_name
        category['id'] = next_category_id
        next_category_id += 1
        category_name_to_category[category_name] = category
    else:
        category = category_name_to_category[category_name]
        assert category['name'] == category_name

    category_id = category['id']

    if category_name == 'unlabeled':
        unlabeled_images.append(im)
    else:
        labeled_images.append(im)
        
    # Create an annotation
    ann = {}
    
    # The Internet tells me this guarantees uniqueness to a reasonable extent, even
    # beyond the sheer improbability of collisions.
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']    
    ann['category_id'] = category_id

    for col in columns_to_copy:
        ann[columns_to_copy[col]] = row[col]
        
    annotations.append(ann)
    
# ...for each image

categories = list(category_name_to_category.values())

elapsed = time.time() - start_time
print('Finished verifying file loop in {}, {} matched images, {} missing images, {} unlabeled images'.format(
        humanfriendly.format_timespan(elapsed), len(images), len(missing_files), len(unlabeled_images))) 


#%% See what's up with missing files

dirnames = set()
# s = list(image_relative_paths)[0]
for s in image_relative_paths:
    image_dir = os.path.dirname(s)
    dirnames.add(image_dir)
        
missing_images_with_missing_dirs = []
missing_images_with_non_missing_dirs = []

missing_dirs = set()

# s = missing_files[0]
for s in missing_files:
    assert s not in image_relative_paths
    dirname = os.path.dirname(s)
    if dirname not in dirnames:
        missing_images_with_missing_dirs.append(s)
        missing_dirs.add(dirname)
    else:
        missing_images_with_non_missing_dirs.append(s)

print('Of {} missing files, {} are due to {} missing folders'.format(
        len(missing_files),len(missing_images_with_missing_dirs),len(missing_dirs)))
    

#%% Check for images that aren't included in the metadata file

unmatched_files = []

for i_image,relative_path in tqdm(enumerate(image_relative_paths),total=len(image_relative_paths)):
    
    if relative_path not in relative_path_to_image:
        unmatched_files.append(relative_path)

print('Finished checking {} images to make sure they\'re in the metadata, found {} mismatches'.format(
        len(image_relative_paths),len(unmatched_files)))


#%% Sample the database

images_all = images
annotations_all = annotations

#%%

if db_sampling_scheme == 'all':
    
    pass

elif db_sampling_scheme == 'labeled' or db_sampling_scheme == 'preview':
    
    json_data = {}
    json_data['images'] = images
    json_data['annotations'] = annotations
    json_data['categories'] = categories

    indexed_db = IndexedJsonDb(json_data)
    
    # Collect the images we want
    sampled_images = []
    for im in images:
        classes = indexed_db.get_classes_for_image(im)
        if 'unlabeled' in classes and len(classes) == 1:
            pass
        else:
            sampled_images.append(im)
    
    if db_sampling_scheme == 'preview':
        n_sample = n_unlabeled_to_sample
        if n_sample == -1:
            n_sample = len(labeled_images)
        if n_sample > len(labeled_images) and cap_unlabeled_to_labeled:
            n_sample = len(labeled_images)
        if n_sample > len(unlabeled_images):
            n_sample = len(unlabeled_images)
        print('Sampling {} of {} unlabeled images'.format(n_sample,len(unlabeled_images)))
        from random import sample
        sampled_images.extend(sample(unlabeled_images,n_sample))

    sampled_annotations = []
    for im in sampled_images:
        sampled_annotations.extend(indexed_db.get_annotations_for_image(im))
    
    print('Sampling {} of {} images, {} of {} annotations'.format(
            len(sampled_images),len(images),len(sampled_annotations),len(annotations)))
  
    images = sampled_images
    annotations = sampled_annotations      
    
else:
    
    raise ValueError('Unrecognized DB sampling scheme {}'.format(db_sampling_scheme))


#%% Create info struct

info = {}
info['year'] = 2019
info['version'] = 1
info['description'] = 'COCO style database'
info['secondary_contributor'] = 'Converted to COCO .json by Dan Morris'
info['contributor'] = 'Parks Canada'


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_file,'w'), indent=4)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate the database's integrity

json_data = json.load(open(output_file))
options = integrity_check_json_db.IntegrityCheckOptions()
sortedCategories,data = integrity_check_json_db.integrity_check_json_db(json_data, options)

    
#%% Render a bunch of images to make sure the labels got carried along correctly

output_dir = preview_base

options = visualize_db.DbVizOptions()
options.num_to_visualize = 100
options.sort_by_filename = False
# options.classes_to_exclude = ['unlabeled']
options.classes_to_exclude = None

htmlOutputFile,_ = visualize_db.visualize_db(json_data,output_dir,input_base,options)
os.startfile(htmlOutputFile)


#%% Write out a list of files to annotate

_,file_list = cct_json_to_filename_json.convertJsonToStringList(output_file,prepend="20190715/")
os.startfile(file_list)




"""

 missouri_to_json.py

 Create .json files from the original source files for the Missouri Camera Traps
 data set.  Metadata was provided here in two formats:

 1) In one subset of the data, folder names indicated species names.  In Set 1,
    there are no empty sequences.  Set 1 has a metadata file to indicate image-level
    bounding boxes.

 2) A subset of the data (overlapping with (1)) was annotated with bounding
    boxes, specified in a whitespace-delimited text file.  In set 2, there are
    some sequences omitted from the metadata file, which implied emptiness.
 
 In the end, set 2 labels were not reliable enough to publish, so LILA includes only set 1.

"""

#%% Constants and imports

import json
import os
import uuid
import time
import humanfriendly
import warnings
import ntpath
import datetime

from PIL import Image

# ignoring all "PIL cannot read EXIF metainfo for the images" warnings
warnings.filterwarnings("ignore", "(Possibly )?corrupt EXIF data", UserWarning)
# Metadata Warning, tag 256 had too many entries: 42, expected 1
warnings.filterwarnings("ignore", "Metadata warning", UserWarning)

baseDir = os.path.expanduser('~/tmp/mct')
imageDir = os.path.join(baseDir,'images')

outputJsonFilenameSet1 = os.path.join(baseDir,'missouri_camera_traps_set1.json')
outputEncoding = 'utf-8'
fileListJsonFilename = os.path.join(baseDir,'images.json')

# This will be a list of filenames that need re-annotation due to redundant boxes
set1RedundantBoxListFilename = os.path.join(baseDir,'mct_images_with_redundant_boxes.txt')

set1BaseDir = os.path.join(imageDir,'Set1')

metadataFilenameSet1 = os.path.join(set1BaseDir,'labels.txt')

assert(os.path.isdir(baseDir))
assert(os.path.isfile(metadataFilenameSet1))

info = {}
info['year'] = 2019
info['version'] = '1.21'
info['description'] = 'Missouri Camera Traps (set 1)'
info['contributor'] = ''
info['date_created'] = str(datetime.date.today())
infoSet1 = info

maxFiles = -1
emptyCategoryId = 0
emptyCategoryName = 'empty'


#%% Enumerate files, read image sizes

# Takes a few minutes, since we're reading image sizes.

# Each element will be a list of relative path/full path/width/height
fileInfo = []
nonImages = []
nFiles = 0

relPathToIm = {}
imageIdToImage = {}

set1ImageIDs = []
   
sequenceIDtoCount = {}

print('Enumerating files from {} to {}'.format(imageDir,fileListJsonFilename))

startTime = time.time()

for root, subdirs, files in os.walk(imageDir):
            
    if root == imageDir:
        continue
    
    bn = ntpath.basename(root)
    
    # Only process leaf nodes corresponding to sequences, which look like:
    #
    # Set1/1.02-Agouti/SEQ75583    
    #
    if ('Set1' in root and 'SEQ' in bn):
        sequenceID = bn
        assert sequenceID not in sequenceIDtoCount
        sequenceIDtoCount[sequenceID] = 0
    else:
        print('Skipping folder {}:{}'.format(root,bn))
        continue
        # assert len(files) <= 2
    
    for fname in files:
  
        nFiles = nFiles + 1
        if maxFiles >= 0 and nFiles > maxFiles:            
            print('Warning: early break at {} files'.format(maxFiles))
            break
        
        fullPath = os.path.join(root,fname)            
        relativePath = os.path.relpath(fullPath,imageDir)
        
        if maxFiles >= 0:
            print(relativePath)
    
        h = -1
        w = -1

        # Read the image
        try:
        
            im = Image.open(fullPath)
            h = im.height
            w = im.width
            
        except:
            
            # Not an image...
            continue
        
        # Store file info
        im = {}
        im['id'] = str(uuid.uuid1())
        im['file_name'] = relativePath
        im['height'] = h
        im['width'] = w
        im['location'] = 'missouri_camera_traps'
        
        im['seq_id'] = sequenceID
        im['seq_num_frames'] = -1
        
        frame_number = sequenceIDtoCount[sequenceID]
        im['frame_num'] = frame_number
        sequenceIDtoCount[sequenceID] = sequenceIDtoCount[sequenceID] + 1
        
        imageIdToImage[im['id']] = im
        relPathToIm[relativePath] = im
        
        if 'Set1' in relativePath:
            set1ImageIDs.append(im['id'])
        else:
            raise Exception('Oops, can''t assign this image to a set')
            
    # ...if we didn't hit the max file limit, keep going
    
    else:
        
        continue
    
    break

# ...for each file

elapsed = time.time() - startTime
print('Finished file enumeration in {}'.format(
      humanfriendly.format_timespan(elapsed)))


#%% Add sequence lengths
    
for imageID in imageIdToImage:
    
    im = imageIdToImage[imageID]
    sequenceID = im['seq_id']
    seq_num_frames = sequenceIDtoCount[sequenceID]
    assert(im['seq_num_frames'] == -1)
    im['seq_num_frames'] = seq_num_frames
    

#%% Load the metadata (.txt) file

with open(metadataFilenameSet1) as f:
    metadataSet1Lines = f.readlines()

metadataSet1Lines = [x.strip() for x in metadataSet1Lines] 


#%% Map relative paths to metadata

# List of lists, length varies according to number of bounding boxes
#
# Preserves original ordering
missingFilesSet1 = []
correctedFiles = []

relPathToMetadataSet1 = {}

# iLine = 0; line = metadataSet1Lines[0]
for iLine,line in enumerate(metadataSet1Lines):
    
    tokens = line.split()
    nTokens = len(tokens)
    
    # Lines should be filename, number of bounding boxes, boxes (four values per box)
    assert ((nTokens - 2) % 4) == 0
    relPath = tokens[0].replace('/',os.sep).replace('\\',os.sep)
    relPath = os.path.join('Set1',relPath)
    absPath = os.path.join(imageDir,relPath)
    
    originalAbsPath = absPath
    originalRelPath = relPath
        
    if not os.path.isfile(absPath):
        
        absPath = originalAbsPath.replace('IMG','IMG_')
        relPath = originalRelPath.replace('IMG','IMG_')
        if os.path.isfile(absPath):
            correctedFiles.append([relPath,originalRelPath,absPath,originalAbsPath])
    
    if not os.path.isfile(absPath):
        
        absPath = originalAbsPath.replace('Red_Deer','Red_Brocket_Deer').replace('IMG','IMG_')
        relPath = originalRelPath.replace('Red_Deer','Red_Brocket_Deer').replace('IMG','IMG_')
        if os.path.isfile(absPath):
            correctedFiles.append([relPath,originalRelPath,absPath,originalAbsPath])
            
    if not os.path.isfile(absPath):
        
        missingFilesSet1.append([originalRelPath,originalAbsPath])
        
    else:
        
        relPathToMetadataSet1[relPath] = tokens
        
        # Make sure we have image info for this image
        assert relPath in relPathToIm

print('Corrected {} paths, missing {} images of {}'.format(len(correctedFiles),
      len(missingFilesSet1),len(metadataSet1Lines)))


#%% Print missing files from Set 1 metadata

# The only missing file (and it's really just missing):
#    
# Set1/1.58-Roe_Deer/SEQ75631/SEQ75631_IMG_0011.JPG

print('Missing files in Set 1:\n')
for iFile,fInfo in enumerate(missingFilesSet1):
    print(fInfo[0])
    

#%% Create categories and annotations for set 1

imagesSet1 = []
categoriesSet1 = []
annotationsSet1 = []

categoryNameToId = {}
idToCategory = {}

# Though we have no empty sequences, we do have empty images in this set
emptyCat = {}
emptyCat['id'] = emptyCategoryId
emptyCat['name'] = emptyCategoryName
emptyCat['count'] = 0
categoriesSet1.append(emptyCat) 

nextCategoryId = emptyCategoryId + 1
    
nFoundMetadata = 0
nTotalBoxes = 0
nImageLevelEmpties = 0
nSequenceLevelAnnotations = 0
nRedundantBoxes = 0

imageIDsWithRedundantBoxes = set()

# For each image
#
# iImage = 0; imageID = set1ImageIDs[iImage]
for iImage,imageID in enumerate(set1ImageIDs):
    
    im = imageIdToImage[imageID]
    imagesSet1.append(im)
    
    # E.g. Set1\\1.80-Coiban_Agouti\\SEQ83155\\SEQ83155_IMG_0010.JPG
    relPath = im['file_name']

    # Find the species name
    tokens = os.path.normpath(relPath).split(os.sep)
    speciesTag = tokens[1]
    tokens = speciesTag.split('-',1)
    assert(len(tokens) == 2)
    categoryName = tokens[1].lower()
    
    category = None
    categoryId = None
    
    if categoryName not in categoryNameToId:
        
        categoryId = nextCategoryId
        nextCategoryId += 1
        categoryNameToId[categoryName] = categoryId
        newCat = {}
        newCat['id'] = categoryNameToId[categoryName]
        newCat['name'] = categoryName
        newCat['count'] = 0
        categoriesSet1.append(newCat) 
        idToCategory[categoryId] = newCat
        category = newCat
        
    else:
        
        categoryId = categoryNameToId[categoryName]
        category = idToCategory[categoryId]
        
        # This image may still be empty...
        # category['count'] = category['count'] + 1
                
    # If we have bounding boxes, create image-level annotations    
    if relPath in relPathToMetadataSet1:
        
        nFoundMetadata += 1
        
        # This tuple is:
        #
        # filename (possibly no longer correct)
        # number of bounding boxes
        # [...boxes (four values per box)]
        imageMetadata = relPathToMetadataSet1[relPath]
        
        nBoxes = int(imageMetadata[1])
        im['n_boxes'] = nBoxes
        
        if nBoxes == 0:
            
            ann = {}
            ann['id'] = str(uuid.uuid1())
            ann['image_id'] = im['id']
            ann['category_id'] = emptyCategoryId
            ann['sequence_level_annotation'] = False
            annotationsSet1.append(ann)
            emptyCat['count'] = emptyCat['count'] + 1
            nImageLevelEmpties += 1
            
        else:
            
            # This image is non-empty
            category['count'] = category['count'] + 1                        
            
            for iBox in range(0,nBoxes):
                                
                boxCoords = imageMetadata[2+(iBox*4):6+(iBox*4)]
                boxCoords = list(map(int, boxCoords))
                
                # Some redundant bounding boxes crept in, don't add them twice
                bRedundantBox = False
                
                # Check this bbox against previous bboxes
                #
                # Inefficient?  Yes.  In an important way?  No.
                for iBoxComparison in range(0,iBox):
                    assert iBox != iBoxComparison                        
                    boxCoordsComparison = imageMetadata[2+(iBoxComparison*4):6+(iBoxComparison*4)]
                    boxCoordsComparison = list(map(int, boxCoordsComparison))
                    if boxCoordsComparison == boxCoords:
                        # print('Warning: redundant box on image {}'.format(relPath))
                        bRedundantBox = True
                        nRedundantBoxes += 1
                        break
                
                if bRedundantBox:
                    imageIDsWithRedundantBoxes.add(im['id'])
                    continue
                    
                # Bounding box values are in absolute coordinates, with the origin 
                # at the upper-left of the image, as [xmin1 ymin1 xmax1 ymax1].
                #
                # Convert to floats and to x/y/w/h, as per CCT standard
                bboxW = boxCoords[2] - boxCoords[0]
                bboxH = boxCoords[3] - boxCoords[1]
                
                box = [boxCoords[0], boxCoords[1], bboxW, bboxH]
                box = list(map(float, box))
                
                ann = {}
                ann['id'] = str(uuid.uuid1())
                ann['image_id'] = im['id']
                ann['category_id'] = categoryId
                ann['sequence_level_annotation'] = False
                ann['bbox'] = box
                annotationsSet1.append(ann)
                nTotalBoxes += 1
            
            # ...for each box
            
        # if we do/don't have boxes for this image
        
    # Else create a sequence-level annotation
    else:
        
        ann = {}
        ann['id'] = str(uuid.uuid1())
        ann['image_id'] = im['id']
        ann['category_id'] = categoryId
        ann['sequence_level_annotation'] = True
        annotationsSet1.append(ann)
        nSequenceLevelAnnotations += 1
        
# ...for each image
        
print('Finished processing set 1, found metadata for {} of {} images'.format(nFoundMetadata,len(set1ImageIDs)))
print('Created {} annotations and {} boxes in {} categories'.format(
    len(annotationsSet1),nTotalBoxes,len(categoriesSet1)))
print('Found {} redundant annotations'.format(nRedundantBoxes))

assert len(annotationsSet1) == nSequenceLevelAnnotations + nTotalBoxes + nImageLevelEmpties
assert len(set1ImageIDs) == nSequenceLevelAnnotations + nFoundMetadata

print('Found {} images with redundant boxes'.format(len(imageIDsWithRedundantBoxes)))


#%% Write out the list of images with redundant boxes

imageFileNamesWithRedundantBoxes = []
for image_id in imageIDsWithRedundantBoxes:
    im = imageIdToImage[image_id]
    imageFileNamesWithRedundantBoxes.append(im['file_name'])
imageFileNamesWithRedundantBoxes.sort()

with open(set1RedundantBoxListFilename,'w') as f:
    for fn in imageFileNamesWithRedundantBoxes:
        f.write(fn + '\n')
    
    
#%% The 'count' field isn't really meaningful, delete it

# It's really the count of image-level annotations, not total images assigned to a class
for d in categoriesSet1:
    del d['count']
    
    
#%% Write output .json files

data = {}
data['info'] = infoSet1
data['images'] = imagesSet1
data['annotations'] = annotationsSet1
data['categories'] = categoriesSet1
json.dump(data, open(outputJsonFilenameSet1,'w'), indent=4)    
print('Finished writing json to {}'.format(outputJsonFilenameSet1))


#%% Consistency-check final set 1 .json file

from megadetector.data_management.databases import integrity_check_json_db
options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = imageDir
options.bCheckImageSizes = True
options.bCheckImageExistence = True
options.bFindUnusedImages = True
options.bRequireLocation = False
options.nThreads = 10
sortedCategories,data,_ = integrity_check_json_db.integrity_check_json_db(outputJsonFilenameSet1, options)
sortedCategories


#%% Generate previews

from megadetector.visualization import visualize_db

output_dir = os.path.join(baseDir,'preview')

options = visualize_db.DbVizOptions()
options.num_to_visualize = 5000
options.sort_by_filename = False
options.classes_to_exclude = None
options.trim_to_images_with_bboxes = False
options.parallelize_rendering = True

htmlOutputFile,_ = visualize_db.visualize_db(outputJsonFilenameSet1,output_dir,imageDir,options)

from megadetector.utils.path_utils import open_file
open_file(htmlOutputFile)
